// SPDX-FileCopyrightText: 2011 - 2022 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "overflowwidget.h"

#include <DGuiApplicationHelper>

#include <QPushButton>
#include <QDebug>
#include <QScrollBar>
#include <QPainterPath>

// Min distance to hide btn
const int SCROLL_MIN_RANGE = 70;
// Once btn pressed, the step of scrollarea
const int SCROLL_BTN_STEP = 50;

// 左右翻页按键大小、左右遮盖大小
const int BTN_SIZE = 24;
const int MASK_SIZE = 28;
const int MASK_MARGIN = 4;

const QString ARROW_UP = QStringLiteral(":/icons/resources/overflow_arrow_up");
const QString ARROW_DOWN = QStringLiteral(":/icons/resources/overflow_arrow_down");
const QString ARROW_LEFT = QStringLiteral(":/icons/resources/overflow_arrow_left");
const QString ARROW_RIGHT = QStringLiteral(":/icons/resources/overflow_arrow_right");

OverflowWidget::OverflowWidget(QWidget *parent)
    : DBlurEffectWidget(parent)
    , m_mainLayout(new QBoxLayout(QBoxLayout::LeftToRight))
    , m_scrollArea(new QScrollArea(this))
    , m_centerScroll(new DBlurEffectWidget(this))
    , m_scrollLayout(new QBoxLayout(QBoxLayout::LeftToRight))
    , m_leftMask(new MaskWidget(this))
    , m_rightMask(new MaskWidget(this))
    , m_leftMaskLayout(new QHBoxLayout)
    , m_rightMaskLayout(new QHBoxLayout)
    , m_leftBtn(new IconButton(this))
    , m_rightBtn(new IconButton(this))
{
    initUI();
    setMaskVisible();
    initConnections();

    m_centerScroll->installEventFilter(this);
    m_scrollArea->installEventFilter(this);
}

void OverflowWidget::initUI()
{
    m_scrollArea->setFrameStyle(QFrame::NoFrame);
    m_scrollArea->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    m_scrollArea->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    m_scrollArea->setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    m_scrollArea->setWidgetResizable(false);
    m_scrollArea->setAutoFillBackground(true);

    m_centerScroll->setLayout(m_scrollLayout);
    m_scrollLayout->setMargin(0);
    m_scrollLayout->setSpacing(0);
    m_centerScroll->setObjectName(OVERFLOWWIDGET_OBJECT_NAME);
    m_centerScroll->setAutoFillBackground(true);
    m_centerScroll->setBlendMode(DBlurEffectWidget::BlendMode::BehindWindowBlend);
    m_scrollArea->setWidget(m_centerScroll);

    m_leftMaskLayout->setContentsMargins(MASK_MARGIN, 0, 0, 0);
    m_leftMaskLayout->setSpacing(0);
    m_leftMaskLayout->addWidget(m_leftBtn, 0, Qt::AlignCenter);
    m_leftMask->setLayout(m_leftMaskLayout);
    m_leftMask->setAutoFillBackground(true);
    m_leftMask->setBlendMode(DBlurEffectWidget::BlendMode::BehindWindowBlend);

    m_rightMaskLayout->setContentsMargins(0, 0, MASK_MARGIN, 0);
    m_rightMaskLayout->setSpacing(0);
    m_rightMaskLayout->addWidget(m_rightBtn, 0, Qt::AlignCenter);
    m_rightMask->setLayout(m_rightMaskLayout);
    m_rightMask->setAutoFillBackground(true);
    m_rightMask->setBlendMode(DBlurEffectWidget::BlendMode::BehindWindowBlend);

    m_leftBtn->setIconPath(ARROW_LEFT);
    m_rightBtn->setIconPath(ARROW_RIGHT);

    m_mainLayout->addWidget(m_leftMask, 0, Qt::AlignCenter);
    m_mainLayout->addWidget(m_scrollArea, 1, Qt::AlignCenter);
    m_mainLayout->addWidget(m_rightMask, 0, Qt::AlignCenter);
    m_mainLayout->setMargin(0);
    m_mainLayout->setSpacing(0);

    setLayout(m_mainLayout);
    setAutoFillBackground(true);
}

void OverflowWidget::initConnections()
{
    connect(m_leftBtn, &QPushButton::clicked, this, &OverflowWidget::leftBtnScrollAction);
    connect(m_rightBtn, &QPushButton::clicked, this, &OverflowWidget::rightBtnScrollAction);
}

void OverflowWidget::setAllMaskAlpha(quint8 alpha)
{
    setMaskAlpha(alpha);
    m_centerScroll->setMaskAlpha(alpha);
    m_leftMask->setMaskAlpha(alpha);
    m_rightMask->setMaskAlpha(alpha);
}

int OverflowWidget::itemCount() const
{
    return m_scrollLayout->count();
}

int OverflowWidget::getJumpStep() const
{
    switch (m_scrollLayout->direction()) {
    case QBoxLayout::LeftToRight:
        return m_scrollArea->width() - SCROLL_BTN_STEP;
    case QBoxLayout::TopToBottom:
        return m_scrollArea->height() - SCROLL_BTN_STEP;
    default:
        return 0;
    }
}

void OverflowWidget::hideItemPopup()
{
    foreach(const auto item, allItem()) {
        if (item->popupShown()) {
            item->hidePopup();
        }
    }
}

void OverflowWidget::leftBtnScrollAction()
{
    int jumpStep = getJumpStep();
    switch (m_scrollLayout->direction()) {
    case QBoxLayout::LeftToRight: {
            int scrollLen = m_scrollArea->horizontalScrollBar()->value() - jumpStep;
            if (scrollLen <= SCROLL_MIN_RANGE) {
                m_scrollArea->horizontalScrollBar()->setValue(0);
            } else {
                m_scrollArea->horizontalScrollBar()->setValue(scrollLen);
            }
        }
        break;
    case QBoxLayout::TopToBottom: {
            int scrollLen = m_scrollArea->verticalScrollBar()->value() - jumpStep;
            if (scrollLen <= SCROLL_MIN_RANGE) {
                m_scrollArea->verticalScrollBar()->setValue(0);
            } else {
                m_scrollArea->verticalScrollBar()->setValue(scrollLen);
            }
        }
        break;
    default:
        break;
    }
    setMaskVisible();
}

void OverflowWidget::rightBtnScrollAction()
{
    int jumpStep = getJumpStep();
    switch (m_scrollLayout->direction()) {
    case QBoxLayout::LeftToRight: {
            int maxLen = m_scrollArea->horizontalScrollBar()->maximum();
            int scrollLen = m_scrollArea->horizontalScrollBar()->value() + jumpStep;
            if (scrollLen > maxLen - SCROLL_MIN_RANGE) {
                m_scrollArea->horizontalScrollBar()->setValue(maxLen);
            } else {
                m_scrollArea->horizontalScrollBar()->setValue(scrollLen);
            }
        }
        break;
    case QBoxLayout::TopToBottom: {
            int maxLen = m_scrollArea->verticalScrollBar()->maximum();
            int scrollLen = m_scrollArea->verticalScrollBar()->value() + jumpStep;
            if (scrollLen > maxLen - SCROLL_MIN_RANGE) {
                m_scrollArea->verticalScrollBar()->setValue(maxLen);
            } else {
                m_scrollArea->verticalScrollBar()->setValue(scrollLen);
            }
        }
        break;
    default:
        break;
    }
    setMaskVisible();
}

void OverflowWidget::setMaskVisible()
{
    bool leftShow = true;
    bool rightShow = true;
    auto hScrollBar = m_scrollArea->horizontalScrollBar();
    auto vScrollBar = m_scrollArea->verticalScrollBar();
    int scrollAreaSize = 0;
    switch (m_scrollLayout->direction()) {
    case QBoxLayout::LeftToRight:
        scrollAreaSize = width() - MASK_SIZE * 2;
        if (hScrollBar->value() <= SCROLL_MIN_RANGE) {
            leftShow = false;
            scrollAreaSize += MASK_SIZE;
        }
        if (hScrollBar->value() >= m_centerScroll->width() - m_scrollArea->width()) {
            rightShow = false;
            scrollAreaSize += MASK_SIZE;
        }
        m_scrollArea->setFixedWidth(scrollAreaSize);
        break;
    case QBoxLayout::TopToBottom:
        scrollAreaSize = height() - MASK_SIZE * 2;
        if (vScrollBar->value() <= SCROLL_MIN_RANGE) {
            leftShow = false;
            scrollAreaSize += MASK_SIZE;
        }
        if (vScrollBar->value() >= m_centerScroll->height() - m_scrollArea->height()) {
            rightShow = false;
            scrollAreaSize += MASK_SIZE;
        }
        m_scrollArea->setFixedHeight(scrollAreaSize);
        break;
    default:
        qWarning() << "The overflow area size is 0 !";
        break;
    }
    m_leftMask->setVisible(leftShow);
    m_rightMask->setVisible(rightShow);
    update();
}

void OverflowWidget::setMaskShape()
{
    switch (m_scrollLayout->direction()) {
    case QBoxLayout::LeftToRight:
        m_leftMask->setFixedSize(MASK_SIZE, height());
        m_rightMask->setFixedSize(MASK_SIZE, height());
        m_leftBtn->setFixedSize(BTN_SIZE, m_leftMask->height() - 8);
        m_rightBtn->setFixedSize(BTN_SIZE, m_rightMask->height() - 8);
        m_leftMaskLayout->setContentsMargins(MASK_MARGIN, 0, 0, 0);
        m_rightMaskLayout->setContentsMargins(0, 0, MASK_MARGIN, 0);
        break;
    case QBoxLayout::TopToBottom:
        m_leftMask->setFixedSize(width(), MASK_SIZE);
        m_rightMask->setFixedSize(width(), MASK_SIZE);
        m_leftBtn->setFixedSize(m_leftMask->width() - 8, BTN_SIZE);
        m_rightBtn->setFixedSize(m_rightMask->width() - 8, BTN_SIZE);
        m_leftMaskLayout->setContentsMargins(0, MASK_MARGIN, 0, 0);
        m_rightMaskLayout->setContentsMargins(0, 0, 0, MASK_MARGIN);
        break;
    default:
        break;
    }
}

void OverflowWidget::setSize(int width, int height, int itemSize)
{
    setFixedSize(width, height);
    m_scrollArea->setFixedSize(width, height);
    const int count = itemCount();
    switch (m_scrollLayout->direction()) {
    case QBoxLayout::TopToBottom:
        m_centerScroll->setFixedSize(width, count * itemSize);
        break;
    case QBoxLayout::LeftToRight:
        m_centerScroll->setFixedSize(count * itemSize, height);
        break;
    default:
        break;
    }
    setMaskShape();
    setMaskVisible();
}

void OverflowWidget::addItem(QWidget *item)
{
    m_scrollLayout->addWidget(item, 0, Qt::AlignCenter);
}

void OverflowWidget::setLayoutPosition(Dock::Position position)
{
    switch (position) {
    case Top:
    case Bottom:
        m_mainLayout->setDirection(QBoxLayout::LeftToRight);
        m_scrollLayout->setDirection(QBoxLayout::LeftToRight);
        m_leftBtn->setIconPath(ARROW_LEFT);
        m_rightBtn->setIconPath(ARROW_RIGHT);
        m_leftMask->setRadiusPos(MaskWidget::RadiusPos::Left);
        m_rightMask->setRadiusPos(MaskWidget::RadiusPos::Right);
        break;
    case Left:
    case Right:
        m_mainLayout->setDirection(QBoxLayout::TopToBottom);
        m_scrollLayout->setDirection(QBoxLayout::TopToBottom);
        m_leftBtn->setIconPath(ARROW_UP);
        m_rightBtn->setIconPath(ARROW_DOWN);
        m_leftMask->setRadiusPos(MaskWidget::RadiusPos::Top);
        m_rightMask->setRadiusPos(MaskWidget::RadiusPos::Bottom);
        break;
    }
    setMaskShape();
    setMaskVisible();
}

QWidget *OverflowWidget::getFirstItem() const
{
    return (m_scrollLayout->count() > 0) ? m_scrollLayout->itemAt(0)->widget() : nullptr;
}

bool OverflowWidget::eventFilter(QObject *watched, QEvent *event)
{
    if (watched != m_centerScroll) {
        return QWidget::eventFilter(watched, event);
    }

    if (event->type() == QEvent::Wheel) {
        hideItemPopup();
        QWheelEvent *wheelEvent = static_cast<QWheelEvent *>(event);
        const QPoint delta = wheelEvent->angleDelta();
        int scroll_len = 0;
        const auto scrollHBar = m_scrollArea->horizontalScrollBar();
        const auto scrollVBar = m_scrollArea->verticalScrollBar();
        if (qAbs(delta.x()) > qAbs(delta.y())) {
            scroll_len = delta.x();
            if (scrollVBar->value() + scroll_len <= SCROLL_MIN_RANGE) {
                scrollVBar->setValue(0);
            } else if (scrollVBar->value() + scroll_len >= scrollVBar->maximum() - SCROLL_MIN_RANGE) {
                scrollVBar->setValue(scrollVBar->maximum());
            } else {
                scrollVBar->setValue(scrollVBar->value() + scroll_len);
            }
        } else {
            scroll_len = (-1 * delta.y());
            if (scrollHBar->value() + scroll_len <= SCROLL_MIN_RANGE) {
                scrollHBar->setValue(0);
            } else if (scrollHBar->value() + scroll_len >= scrollHBar->maximum() - SCROLL_MIN_RANGE) {
                scrollHBar->setValue(scrollHBar->maximum());
            } else {
                scrollHBar->setValue(scrollHBar->value() + scroll_len);
            }
        }
    } else if (event->type() == QEvent::Resize) {
        m_scrollArea->horizontalScrollBar()->update();
        m_scrollArea->verticalScrollBar()->update();
    } else {
        return QWidget::eventFilter(watched, event);
    }

    setMaskVisible();
    return false;
}

void OverflowWidget::setAllRadius(int radius)
{
    if (!DWindowManagerHelper::instance()->hasComposite()) {
        radius = 0;
    }

    setRadius(radius);
    setBlurRectXRadius(radius);
    setBlurRectYRadius(radius);

    m_centerScroll->setRadius(radius);
    m_centerScroll->setBlurRectXRadius(radius);
    m_centerScroll->setBlurRectYRadius(radius);

    m_leftMask->setRadius(radius);
    m_leftMask->setBlurRectXRadius(radius);
    m_leftMask->setBlurRectYRadius(radius);

    m_rightMask->setRadius(radius);
    m_rightMask->setBlurRectXRadius(radius);
    m_rightMask->setBlurRectYRadius(radius);

    update();
}

void MaskWidget::setRadiusPos(const RadiusPos &pos)
{
    m_radiusPos = pos;
    update();
}

void MaskWidget::paintEvent(QPaintEvent *e)
{
    QPainter painter(this);

    QPainterPath path;
    path.setFillRule( Qt::WindingFill );
    path.addRoundedRect (rect(), radius(), radius());
    switch ( m_radiusPos ) {
    case RadiusPos::Left:
        path.addRect(QRect(rect().left()+radius(), rect().top(), rect().width()-radius(), rect().height()));
        break;
    case RadiusPos::Right:
        path.addRect(QRect(rect().left(), rect().top(), rect().width()-radius(), rect().height()));
        break;
    case RadiusPos::Top:
        path.addRect(QRect(rect().left(), rect().top()+radius(), rect().width(), rect().height()-radius()));
        break;
    case RadiusPos::Bottom:
        path.addRect(QRect(rect().left(), rect().top(), rect().width(), rect().height()-radius()));
        break;
    default:
        path.addRect(QRect(rect().left()+radius(), rect().top(), rect().width()-radius(), rect().height()));
        break;
    }
    painter.setRenderHint(QPainter::Antialiasing);
    painter.setClipPath(path);

    painter.fillRect(rect(), maskColor());
}

IconButton::IconButton(QWidget *parent)
    : QPushButton(parent)
{
    setFocusPolicy(Qt::NoFocus);
    setFlat(true);
}

void IconButton::paintEvent(QPaintEvent *e)
{
    QPainter painter(this);
    painter.setRenderHint(QPainter::Antialiasing, true);
    painter.setRenderHint(QPainter::SmoothPixmapTransform, true);

    QColor color;
    if (DGuiApplicationHelper::instance()->themeType() == DGuiApplicationHelper::LightType) {
        color = Qt::black;
        color.setAlphaF(0.2);
    } else {
        color = QColor("#CDCDCD");
        color.setAlphaF(0.1);
    }
    QPainterPath path;
    path.addRoundedRect(rect(), 8, 8);
    painter.fillPath(path, color);

    if (!icon().isNull()) {
        const auto ratio = devicePixelRatioF();
        const QSize pixmapSize = QCoreApplication::testAttribute(Qt::AA_UseHighDpiPixmaps) ? iconSize() : (iconSize() * ratio);
        QPixmap pixmap = icon().pixmap(pixmapSize);
        pixmap.setDevicePixelRatio(ratio);
        const QRectF &rf = QRectF(rect());
        const QRectF &rfp = QRectF(pixmap.rect());
        painter.drawPixmap(rf.center() - rfp.center() / pixmap.devicePixelRatioF(), pixmap);
    }
}
