// SPDX-FileCopyrightText: 2011 - 2022 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#ifndef DOCKITEM_H
#define DOCKITEM_H

#include "constants.h"
#include "dockpopupwindow.h"

#include <QFrame>
#include <QPointer>
#include <QGestureEvent>
#include <QMenu>

#include <memory>

const QString OVERFLOWWIDGET_OBJECT_NAME = QStringLiteral("overflowScroll");

using namespace Dock;

class DockItemContextMenu : public QMenu
{
public:
    using QMenu::QMenu;

    void clearShowReminderActions()
    {
        m_showReminderActions.clear();
    }

    void addShowReminderAction(QAction* action)
    {
        m_showReminderActions.append(action);
    }

    // action宽度的计算方法：文字宽度 + 小红点（如果有的话）+ 左右边距60
    // 菜单的宽度为action的最大宽度与160（设计的最小宽度）两者取较大值
    int suitableWidth() const
    {
        int maxWidth = 0;
        for (auto action : actions()) {
            auto textWidth = fontMetrics().width(action->text());
            textWidth += m_showReminderActions.contains(action) ? 26 : 0;
            maxWidth = maxWidth < textWidth ? textWidth : maxWidth;
        }

        return qMax(maxWidth + 60, 160);
    }

protected:
    void paintEvent(QPaintEvent *e) override
    {
        QMenu::paintEvent(e);

        QPainter p(this);
        p.setRenderHint(QPainter::Antialiasing, true);
        for (auto action : m_showReminderActions) {
            auto geo = actionGeometry(action);
            QColor color("#FF3B30");
            p.setPen(color);
            p.setBrush(color);
            p.drawEllipse(geo.x() + geo.width() - 26, geo.y() + (geo.height() - 6) / 2, 6, 6);
        }

        p.end();
    }

private:
    QList<QPointer<QAction>> m_showReminderActions;
};

class DockItem : public QWidget
{
    Q_OBJECT

public:
    enum ItemType {
        Launcher,
        App,
        Plugins,
        FixedPlugin,
        Placeholder,
        Overflow,
        TrayPlugin,
    };

public:
    explicit DockItem(QWidget *parent = nullptr);
    ~DockItem() override;

    static void setDockPosition(Position side);
    static void setDockSize(int size);
    static void setDockDisplayMode(DisplayMode mode);

    inline virtual ItemType itemType() const { return App; }

    QSize sizeHint() const override;
    virtual QString accessibleName();
    inline bool popupShown() const { return m_popupShown; }

public slots:

    virtual void refreshIcon() { }

    void showPopupApplet(QWidget *applet);
    void hidePopup();
    virtual void setDragging(bool bDrag);

    virtual void checkEntry() { }

    bool isDragging() const;
signals:
    void dragStarted() const;
    void itemDropped(QObject *destination, const QPoint &dropPoint) const;
    void requestWindowAutoHide(const bool autoHide) const;
    void requestRefreshWindowVisible() const;
    void hideOverflowPopup() const;

protected:
    bool event(QEvent *event) override;
    void paintEvent(QPaintEvent *e) override;
    void mousePressEvent(QMouseEvent *e) override;
    void contextMenuEvent(QContextMenuEvent *e) override;
    void enterEvent(QEvent *e) override;
    void leaveEvent(QEvent *e) override;

    const QRect perfectIconRect() const;
    const QPoint popupMarkPoint();
    QPoint topLeftPoint() const;

    void hideNonModel();
    void popupWindowAccept();
    virtual void showPopupWindow(QWidget *content, bool model = false, int radius = 6);
    virtual void showHoverTips();
    virtual void invokedMenuItem(const QString &itemId, bool checked);
    virtual const QString contextMenu() const;
    virtual QWidget *popupTips();

    bool checkAndResetTapHoldGestureState();
    virtual void gestureEvent(QGestureEvent *event);

protected slots:
    void showContextMenu();
    void onContextMenuAccepted() const;

private:
    void updatePopupPosition();
    void menuActionClicked(QAction *action);

protected:
    bool m_hover;
    bool m_popupShown;
    bool m_tapAndHold;
    bool m_draging;
    DockItemContextMenu m_contextMenu;

    DWindowManagerHelper *m_wmHelper;

    QPointer<QWidget> m_lastPopupWidget;

    QTimer *m_popupTipsDelayTimer;
    QTimer *m_popupAdjustDelayTimer;

    static Position DockPosition;
    static int DockSize;
    static DisplayMode DockDisplayMode;
    static QPointer<DockPopupWindow> PopupWindow;
};

#endif // DOCKITEM_H
