// SPDX-FileCopyrightText: 2011 - 2022 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "overflowitem.h"
#include "diconbutton.h"
#include "appitem.h"
#include "themeappicon.h"

#include <DGuiApplicationHelper>
#include <DPaletteHelper>
#include <DPalette>
#include <DHiDPIHelper>

#include <QBoxLayout>
#include <QLabel>
#include <QScrollArea>
#include <QScrollBar>

const QString ICON_DEFAULT = QStringLiteral(":/icons/resources/application-x-desktop");
const QString OVERFLOW_MORE = QStringLiteral(":/icons/resources/overflow-more");

// 浅色和深色主题下背景阴影透明度
const QMap<QEvent::Type, double> SHADOW_OPACITY_LIGHT = { {QEvent::Type::Enter, 0.3},
                                                          {QEvent::Type::Leave, 0.2},
                                                          {QEvent::Type::MouseButtonPress, 0.4},
                                                          {QEvent::Type::MouseButtonRelease, 0.2} };
const QMap<QEvent::Type, double> SHADOW_OPACITY_DARK = { {QEvent::Type::Enter, 0.3},
                                                         {QEvent::Type::Leave, 0.2},
                                                         {QEvent::Type::MouseButtonPress, 0.1},
                                                         {QEvent::Type::MouseButtonRelease, 0.2} };

// PopWindow distance Dock
const int POPUP_DOCK_DISTANCE = 10;

// 溢出区弹窗圆角
const int POPUP_RADIUS = 10;

// check if is darktheme
inline bool isDarkTheme()
{
    return DGuiApplicationHelper::instance()->themeType() == DGuiApplicationHelper::DarkType;
}

OverflowItem::OverflowItem(QWidget *parent)
    : DockItem(parent)
    , m_showPopup(false)
    , m_overflowIcon(ICON_DEFAULT)
    , m_moreIcon(QPixmap(OVERFLOW_MORE))
    , m_popupMargin(10)
    , m_popupWindow(new DockPopupWindow(nullptr, true, true))
    , m_contentWidget(new OverflowWidget)
    , m_regionInter(new DRegionMonitor(this))
    , m_tips(new TipsWidget(this))
    , m_state(QEvent::Type::Leave)
{
    initUI();
    initConnections();
}

OverflowItem::~OverflowItem()
{
    m_popupWindow->hide();
    m_popupWindow->deleteLater();
    m_contentWidget->deleteLater();
}

void OverflowItem::initUI()
{
    m_tips->setText(tr("Show hidden icons"));
    m_tips->setObjectName("HiddenOverflowTip");
    m_tips->setVisible(false);

    m_popupWindow->setShadowBlurRadius(0);
    m_popupWindow->setRadius(0);
    m_popupWindow->setShadowYOffset(0);
    m_popupWindow->setShadowXOffset(0);
    m_popupWindow->setArrowWidth(0);
    m_popupWindow->setArrowHeight(0);
    m_popupWindow->setObjectName("overflowpopup");
    m_popupWindow->setLeftRightRadius(true);
    m_popupWindow->installEventFilter(this);
}

void OverflowItem::initConnections()
{    // listen to the region signal
    connect(m_regionInter, &DRegionMonitor::buttonRelease, this, &OverflowItem::onGlobMouseRelease);
    connect(this, &OverflowItem::maskAlphaChanged, m_contentWidget, &OverflowWidget::setAllMaskAlpha, Qt::QueuedConnection);
}

int OverflowItem::itemCountOnZone() const
{
    return m_contentWidget->itemCount();
}

void OverflowItem::refreshIcon()
{
    QWidget *widget = m_contentWidget->getFirstItem();
    if (widget) {
        const QString appIconName = qobject_cast<AppItem *>(widget)->appIconName();

        if (appIconName.isNull()) {
            return;
        }

        if(!ThemeAppIcon::getIcon(m_overflowIcon, appIconName, size().width() * 0.7, false)) {
            m_overflowIcon = QPixmap(ICON_DEFAULT);
        }

        // 应用icon降为不透明度30%
        QPixmap resultIcon = m_overflowIcon;
        QPainter painter(&resultIcon);
        painter.setCompositionMode(QPainter::CompositionMode_Source);
        painter.fillRect(resultIcon.rect(), Qt::transparent);
        painter.setCompositionMode(QPainter::CompositionMode_SourceOver);
        painter.drawPixmap(0, 0, m_overflowIcon);
        painter.setCompositionMode(QPainter::CompositionMode_DestinationIn);
        painter.fillRect(resultIcon.rect(), QColor(0, 0, 0, 0.3 * 255));
        painter.end();
        m_overflowIcon = resultIcon;
    }
    update();
}

void OverflowItem::hidePopUpWindow()
{
    if (m_regionInter->registered())
        m_regionInter->unregisterRegion();
    m_showPopup = false;
    m_popupWindow->hide();

    // 隐藏后，不响应tip
    setItemTipSwitch(false);

    emit popupWindowShowChanged(false, QRect());

    // 等待popup完全隐藏后，更新溢出区应用的最小化位置，从本来的位置，变为溢出区按键位置
    QTimer::singleShot(200, this, [ = ] {
        updateItemMinimizedGeometry(true);
    });
}

void OverflowItem::setItemTipSwitch(bool needTip)
{
    const QList<DockItem *> list = m_contentWidget->allItem();
    for (const auto item : list) {
        qobject_cast<AppItem *>(item)->setTipSwitch(needTip);
    }
}

void OverflowItem::updateItemMinimizedGeometry(bool needUpdate)
{
    const QList<DockItem *> list = m_contentWidget->allItem();
    for (const auto item : list) {
        qobject_cast<AppItem *>(item)->updateOverflowAppMinimizedGeometry(
            true,
            needUpdate,
            QRect(mapToGlobal(QPoint(0, 0)), mapToGlobal(QPoint(width(), height())))
            );
    }
}

void OverflowItem::setPopUpSize(int width, int height, QSize panelSize, QPoint panelPos, int itemSize)
{
    m_contentWidget->setSize(width, height, itemSize);
    m_panelSize = panelSize;
    m_panelPos = panelPos;
}

void OverflowItem::addItem(QWidget *item)
{
    m_contentWidget->addItem(item);
}

QPoint OverflowItem::overflowIconPosition(const QPixmap &pixmap) const
{
    const auto ratio = devicePixelRatioF();
    const QRectF itemRect = rect();
    const QRectF iconRect = pixmap.rect();
    const qreal iconX = itemRect.center().x() - iconRect.center().x() / ratio;
    const qreal iconY = itemRect.center().y() - iconRect.center().y() / ratio;
    return QPoint(iconX, iconY);
}

void OverflowItem::paintEvent(QPaintEvent *e)
{
    DockItem::paintEvent(e);

    QPainter painter(this);

    // set paint hint
    painter.setRenderHint(QPainter::Antialiasing);
    painter.setRenderHint(QPainter::SmoothPixmapTransform);

    // Start paint image
    QPoint realPos = overflowIconPosition(m_overflowIcon);
    painter.drawPixmap(realPos, m_overflowIcon);
    // Paint End

    // Add Shadow
    QColor shadowColor;
    if (isDarkTheme()) {
        shadowColor = Qt::white;
        painter.setOpacity(SHADOW_OPACITY_DARK.value(m_state));
    } else {
        shadowColor = Qt::black;
        painter.setOpacity(SHADOW_OPACITY_LIGHT.value(m_state));
    }

    const QRectF itemRect = rect();
    const qreal min = qMin(itemRect.width(), itemRect.height());
    QRectF backgroundRect = QRectF(itemRect.x(), itemRect.y(), min, min);
    backgroundRect = backgroundRect.marginsRemoved(QMargins(2, 2, 2, 2));
    backgroundRect.moveCenter(itemRect.center());

    QPainterPath path;
    path.addRoundedRect(backgroundRect, 8, 8);
    painter.fillPath(path, shadowColor);
    // Shadow end

    // Add More Icon
    painter.setOpacity(1);
    m_moreIcon.setDevicePixelRatio(devicePixelRatioF());
    painter.drawPixmap(overflowIconPosition(m_moreIcon), m_moreIcon);
    // Paint "More" End
}

// INFO: public, be set in mainpanelcontrol.cpp, not use it in this cpp
void OverflowItem::setLayoutPosition(Dock::Position position)
{
    m_contentWidget->setLayoutPosition(position);
}

void OverflowItem::mousePressEvent(QMouseEvent *e)
{
    if (e->button() != Qt::LeftButton) {
        DockItem::mousePressEvent(e);
        return;
    }

    m_showPopup = !m_showPopup;
    if (m_showPopup) {
        m_popupWindow->setLeftRightRadius(DWindowManagerHelper::instance()->hasComposite());
        m_contentWidget->setAllRadius(POPUP_RADIUS);
        // INFO: 因为抢占后端事件，如果托盘点击后，没有取消注册，那么这个地方就无法注册上
        // 理论应该是dtkgui或者后端的bug,registered 函数应该返回一个bool类型，告知是否注册
        // 如果有这种行为那么应该处理为无法注册就不弹出窗口
        // 现在只能延迟显示，等待托盘解除注册事件
        QTimer::singleShot(150, this, [this]{
            showOverflowPopupWindow(m_contentWidget, true, POPUP_RADIUS);
        });
    } else {
        hidePopUpWindow();
    }
    DockItem::mousePressEvent(e);
}

void OverflowItem::showOverflowPopupWindow(QWidget *const content, const bool model, const int radius)
{
    m_popupShown = true;
    m_lastPopupWidget = content;

    if (model)
        emit requestWindowAutoHide(false);

    m_popupWindow->setRadius(radius);
    QWidget *lastContent = m_popupWindow->getContent();
    if (lastContent)
        lastContent->setVisible(false);

    m_popupWindow->resize(content->sizeHint());
    m_popupWindow->setContent(content);

    // 显示时，响应tip
    setItemTipSwitch(true);

    updateItemMinimizedGeometry(false);

    if (!m_popupWindow->isVisible())
        QMetaObject::invokeMethod(m_popupWindow, "show", Qt::QueuedConnection, Q_ARG(QPoint, calcPopupWindowCenterPos()), Q_ARG(bool, model));
    else
        m_popupWindow->show(calcPopupWindowCenterPos(), model);
    m_regionInter->registerRegion();

    emit popupWindowShowChanged(true, calcPopupWindowRect());
}

bool OverflowItem::event(QEvent *event)
{
    switch(event->type()) {
    case QEvent::Enter:
    case QEvent::Leave:
    case QEvent::MouseButtonPress:
    case QEvent::MouseButtonRelease:
        m_state = event->type();
        update();
        break;

    case QEvent::Resize:
        m_moreIcon = DHiDPIHelper::loadNxPixmap(OVERFLOW_MORE).scaled(size() * 0.6, Qt::KeepAspectRatio, Qt::SmoothTransformation);
        update();
        break;

    default:
        break;
    }

    return DockItem::event(event);
}

bool OverflowItem::eventFilter(QObject *watched, QEvent *event)
{
    if ((watched == m_popupWindow) && (event->type() == QEvent::Leave)) {
        emit popupWindowCursorOut();
    }
    return DockItem::eventFilter(watched, event);
}

QPoint OverflowItem::calcPopupWindowCenterPos()
{
    QPoint position = popupMarkPoint();
    const QSize contentSize = m_contentWidget->size();
    auto posAdjustLeftToRight = [contentSize, this, &position] {
        // 补偿点，当双屏且水平排列时，任务栏在底/顶部的起始点不是0
        QPoint compensate(0, 0);
        compensate.setX(m_panelPos.x());
        position.rx() = position.x() - m_panelPos.x();

        if (position.x() + contentSize.width() / 2 >= m_panelSize.width() - m_popupMargin * 2) {
            if (contentSize.width() >= m_panelSize.width() - m_popupMargin * 2) {
                // 以屏幕中间点为中心
                position.rx() = m_panelSize.width() / 2 ;
            } else {
                // 中心点由溢出区按键逐渐向屏幕中间点靠拢
                position.rx() = m_panelSize.width() - contentSize.width() / 2 - m_popupMargin;
            }
        } else if (position.x() < contentSize.width() / 2) {
            // 以溢出区按键为中心点
            position.rx() = contentSize.width() / 2 + m_popupMargin;
        }
        position.rx() = position.x() + compensate.x();
    };
    // 同上
    auto posAdjustTopToBottom = [contentSize, this, &position] {
        QPoint compensate(0, 0);
        compensate.setY(m_panelPos.y());
        position.ry() = position.y() - m_panelPos.y();

        if (position.y() + contentSize.height() / 2 >= m_panelSize.height() - m_popupMargin * 2) {
            if (contentSize.height() >= m_panelSize.height() - m_popupMargin * 2) {
                position.ry() = m_panelSize.height() / 2;
            } else {
                position.ry() = m_panelSize.height() - contentSize.height() / 2 - m_popupMargin;
            }
        } else if (position.y() < contentSize.height() / 2) {
            position.ry() = contentSize.height() / 2 + m_popupMargin;
        }
        position.ry() = position.y() + compensate.y();
    };

    switch (DockPosition) {
    case Top:
        m_popupWindow->setArrowDirection(DockPopupWindow::ArrowTop);
        position.ry() += POPUP_DOCK_DISTANCE;
        posAdjustLeftToRight();
        break;
    case Bottom:
        m_popupWindow->setArrowDirection(DockPopupWindow::ArrowBottom);
        position.ry() -= POPUP_DOCK_DISTANCE;
        posAdjustLeftToRight();
        break;
    case Left:
        m_popupWindow->setArrowDirection(DockPopupWindow::ArrowLeft);
        position.rx() += POPUP_DOCK_DISTANCE;
        posAdjustTopToBottom();
        break;
    case Right:
        m_popupWindow->setArrowDirection(DockPopupWindow::ArrowRight);
        position.rx() -= POPUP_DOCK_DISTANCE;
        posAdjustTopToBottom();
        break;
    }
    return  position;
}

QRect OverflowItem::calcPopupWindowRect()
{
    QRect rect;
    const QPoint popupCenterPos = calcPopupWindowCenterPos();
    const QSize contentSize = m_contentWidget->size();
    switch (DockPosition) {
    case Top:
        rect.setX(popupCenterPos.x() - (contentSize.width() / 2));
        rect.setY(popupCenterPos.y());
        break;
    case Bottom:
        rect.setX(popupCenterPos.x() - (contentSize.width() / 2));
        rect.setY(popupCenterPos.y() - contentSize.height());
        break;
    case Left:
        rect.setX(popupCenterPos.x());
        rect.setY(popupCenterPos.y() - (contentSize.height() / 2));
        break;
    case Right:
        rect.setX(popupCenterPos.x() - contentSize.width());
        rect.setY(popupCenterPos.y() - (contentSize.height() / 2));
        break;
    }
    rect.setWidth(contentSize.width());
    rect.setHeight(contentSize.height());
    return rect;
}

void OverflowItem::onGlobMouseRelease(const QPoint &mousePos, const int flag)
{
    if ((flag != DRegionMonitor::WatchedFlags::Button_Left)
       && (flag != DRegionMonitor::WatchedFlags::Button_Right)) {
        return;
    }

    auto position = pos();
    const QWidget *w = qobject_cast<QWidget *>(this->parent());
    while (w) {
        position += w->pos();
        w = qobject_cast<QWidget *>(w->parent());
    }
    QRect itemRect = rect();
    itemRect.moveTo(position);

    if (itemRect.contains(mousePos))
        return;

    const QRect popRect = QRect(m_popupWindow->pos(), m_popupWindow->size());
    if (popRect.contains(mousePos))
        return;

    hidePopUpWindow();
};

QWidget *OverflowItem::popupTips()
{
    if (m_popupWindow->isVisible()) {
        return nullptr;
    }

    return m_tips;
}

void OverflowItem::setLayoutMargin(int margin)
{
    m_popupMargin = margin;
}
