// SPDX-FileCopyrightText: 2011 - 2022 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "mainwindow.h"
#include "accessible.h"
#include "dbusdockadaptors.h"
#include "utils.h"
#include "themeappicon.h"
#include "dockitemmanager.h"
#include "dockapplication.h"
#include "settingmanager.h"

#include <QAccessible>
#include <QDir>
#include <QStandardPaths>
#include <QDateTime>
#include <QDir>

#include <DApplication>
#include <DLog>
#include <DGuiApplicationHelper>

#include <unistd.h>
#include <string>
#include <sys/mman.h>
#include <stdio.h>
#include <time.h>
#include <execinfo.h>
#include <sys/stat.h>

DWIDGET_USE_NAMESPACE
#ifdef DCORE_NAMESPACE
DCORE_USE_NAMESPACE
#else
DUTIL_USE_NAMESPACE
#endif

// 用于在脚本中判断任务栏是否异常退出，如果是1也认为是正常退出，不计入异常退出次数
#define EXIT_BY_MYSELF (1)

int main(int argc, char *argv[])
{
    if (QString(getenv("XDG_CURRENT_DESKTOP")).compare("deepin", Qt::CaseInsensitive) == 0) {
        qDebug() << "Warning: force enable D_DXCB_FORCE_NO_TITLEBAR now!";
        setenv("D_DXCB_FORCE_NO_TITLEBAR", "1", 1);
    }

    DGuiApplicationHelper::setAttribute(DGuiApplicationHelper::UseInactiveColorGroup, false);
    DockApplication app(argc, argv);
    app.setOrganizationName("deepin");
    app.setApplicationName("dde-dock");
    app.setApplicationDisplayName("DDE Dock");
    app.setApplicationVersion("2.0");
    app.loadTranslator();
    app.setAttribute(Qt::AA_EnableHighDpiScaling, true);
    app.setAttribute(Qt::AA_UseHighDpiPixmaps, true);

    // 自动化标记
    QAccessible::installFactory(accessibleFactory);

    // 设置日志输出到控制台以及文件
    DLogManager::registerConsoleAppender();
    DLogManager::registerFileAppender();

    // 启动入参 dde-dock --help可以看到一下内容， -x不加载插件 -r 一般用在startdde启动任务栏
    QCommandLineOption disablePlugOption(QStringList() << "x" << "disable-plugins", "do not load plugins.");
    QCommandLineOption runOption(QStringList() << "r" << "run-by-startdde", "run by startdde.");
    QCommandLineParser parser;
    parser.setApplicationDescription("DDE Dock");
    parser.addHelpOption();
    parser.addVersionOption();
    parser.addOption(disablePlugOption);
    parser.addOption(runOption);
    parser.process(app);

    // 任务栏单进程限制
    DGuiApplicationHelper::setSingleInstanceInterval(-1);
    if (!app.setSingleInstance(QString("dde-dock_%1").arg(getuid()))) {
        qDebug() << "set single instance failed!";
        return EXIT_BY_MYSELF;
    }

#ifndef QT_DEBUG
    QDir::setCurrent(QApplication::applicationDirPath());
#endif

    // 注册任务栏的DBus服务
    MainWindow mw;
    DBusDockAdaptors adaptor(&mw);

    if(Utils::IS_WAYLAND_DISPLAY) {
        mw.setAttribute(Qt::WA_NativeWindow);
        mw.windowHandle()->setProperty("_d_dwayland_window-type", "dock");
    }

    QDBusConnection::sessionBus().registerService("com.deepin.dde.Dock");
    QDBusConnection::sessionBus().registerObject("/com/deepin/dde/Dock", "com.deepin.dde.Dock", &mw);

    // 当任务栏以-r参数启动时，设置CANSHOW未false，之后调用launch不显示任务栏
    qApp->setProperty("CANSHOW", !parser.isSet(runOption));

    mw.launch();

    /*
     * 之所以这样写，是为了在未设置 -x 运行参数时，不用花费时间去获取 enableSafeMode 的配置，可以加快一些启动速度
     */
    if (parser.isSet(disablePlugOption)) {
        // 是否启动安全模式的功能，启用后，如果频繁异常退出则会自动进入安全模式
        if (SettingManager::instance()->enableSafeMode()) {
            mw.sendNotifications();
        } else {
            goto continue_load;
        }
    } else {
        goto continue_load;
    }

    return app.exec();

continue_load:
    DockItemManager::instance()->startLoadPlugins();
    qApp->setProperty("PLUGINSLOADED", true);

    // 以前的一些旧的配置文件，现在用不到了,稍后删除同样是为了加快主要的任务
    QMetaObject::invokeMethod(qApp, []{
        const QString &cfgPath = QStandardPaths::standardLocations(QStandardPaths::ConfigLocation)[0] + "/dde-cfg.ini";
        if (QFile::exists(cfgPath))
            QFile::remove(cfgPath);

        QDir dir(QStandardPaths::standardLocations(QStandardPaths::CacheLocation)[0]);
        dir.cdUp();
        const QString &collapseFile = dir.path() + "/dde-collapse.log";
        if (QFile::exists(collapseFile))
            QFile::remove(collapseFile);
    }, Qt::QueuedConnection);

    return app.exec();
}
