// SPDX-FileCopyrightText: 2011 - 2022 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "dockpopupwindow.h"
#include "utils.h"

#include <DForeignWindow>

#include <QAccessible>
#include <QMetaObject>
#include <qpa/qplatformwindow.h>

DWIDGET_USE_NAMESPACE

DockPopupWindow::DockPopupWindow(QWidget *parent, bool noFocus, bool tellWmDock)
    : DArrowRectangle(ArrowBottom, parent),
      m_model(false),
      m_regionInter(new DRegionMonitor(this))
{
    setMargin(0);
    m_wmHelper = DWindowManagerHelper::instance();

    compositeChanged();
    setWindowFlag(Qt::X11BypassWindowManagerHint);
    if (Utils::IS_WAYLAND_DISPLAY) {
        setAttribute(Qt::WA_NativeWindow);
    } else {
        setAttribute(Qt::WA_InputMethodEnabled, false);
    }
    if (noFocus) {
        setWindowFlag(Qt::WindowDoesNotAcceptFocus);
        setWindowFlag(Qt::WindowStaysOnTopHint);
    }
    if (Utils::IS_WAYLAND_DISPLAY && tellWmDock) {
         // 谨慎修改层级，特别要注意对锁屏的影响
         windowHandle()->setProperty("_d_dwayland_window-type", "dock");
    }

    // Fix: bug-211649 在该对象显示前调用DWindowManagerHelper::currentWorkspaceWindows会造成背景颜色异常，提前创建winId规避这个问题
    createWinId();

    connect(m_wmHelper, &DWindowManagerHelper::hasCompositeChanged, this, &DockPopupWindow::compositeChanged);
    connect(m_regionInter, &DRegionMonitor::buttonPress, this, &DockPopupWindow::onGlobMousePress);
}

DockPopupWindow::~DockPopupWindow()
{
}

bool DockPopupWindow::model() const
{
    return isVisible() && m_model;
}

void DockPopupWindow::setContent(QWidget *content)
{
    QWidget *lastWidget = getContent();
    if (lastWidget)
        lastWidget->removeEventFilter(this);
    content->installEventFilter(this);

    QAccessibleEvent event(this, QAccessible::NameChanged);
    QAccessible::updateAccessibility(&event);

    if (!content->objectName().trimmed().isEmpty())
        setAccessibleName(content->objectName() + "-popup");

    DArrowRectangle::setContent(content);
}

void DockPopupWindow::show(const QPoint &pos, const bool model)
{
    m_model = model;
    m_lastPoint = pos;

    show(pos.x(), pos.y());

    if (m_regionInter->registered()) {
        m_regionInter->unregisterRegion();
    }

    if (m_model) {
        // 此处需要延迟注册，因为在wayland下，这里如果马上注册就会立刻触发m_regionInter的buttonPress信号（具体原因需要DRegionMonitor来分析）
        // 导致出现一系列问题，延迟注册后就不会触发buttonPress信号，保证流程正确
        QMetaObject::invokeMethod(m_regionInter, "registerRegion", Qt::QueuedConnection);
    }
}

void DockPopupWindow::show(const int x, const int y)
{
    m_lastPoint = QPoint(x, y);

    DArrowRectangle::show(x, y);
}

void DockPopupWindow::hide()
{
    if (m_regionInter->registered())
        m_regionInter->unregisterRegion();

    DArrowRectangle::hide();
}

void DockPopupWindow::showEvent(QShowEvent *e)
{
    DArrowRectangle::showEvent(e);
    if (Utils::IS_WAYLAND_DISPLAY) {
        Utils::updateCursor(this);
    }

    QTimer::singleShot(1, this, &DockPopupWindow::ensureRaised);
}

void DockPopupWindow::enterEvent(QEvent *e)
{
    DArrowRectangle::enterEvent(e);
    if (Utils::IS_WAYLAND_DISPLAY) {
        Utils::updateCursor(this);
    }

    QTimer::singleShot(1, this, &DockPopupWindow::ensureRaised);
}

bool DockPopupWindow::eventFilter(QObject *o, QEvent *e)
{
    if (o != getContent() || e->type() != QEvent::Resize)
        return false;

    // FIXME: ensure position move after global mouse release event
    if (isVisible()) {
        QTimer::singleShot(10, this, [=] {
            // NOTE(sbw): double check is necessary, in this time, the popup maybe already hided.
            if (isVisible())
                show(m_lastPoint, m_model);
        });
    }

    return false;
}

void DockPopupWindow::onGlobMousePress(const QPoint &mousePos, const int flag)
{
    Q_ASSERT(m_model);

    if (!((flag == DRegionMonitor::WatchedFlags::Button_Left) ||
          (flag == DRegionMonitor::WatchedFlags::Button_Right))) {
        return;
    }

    const QRect rect = QRect(pos(), size());
    if (rect.contains(mousePos)) {
        // 如果此时窗口没有激活，手动激活一下，否则用户无法输入
        if (!isActiveWindow() && !windowFlags().testFlag(Qt::WindowDoesNotAcceptFocus)) {
            activateWindow();
        }
        return;
    }

    // 如果点击的是屏幕键盘，则不隐藏
    QList<Dtk::Gui::DForeignWindow*> windowList = m_wmHelper->currentWorkspaceWindows();
    for (auto window : windowList) {
        if (window->wmClass() == "onboard" && window->handle()->geometry().contains(scalePoint(mousePos)))
            return;
    }

    emit accept();

    // 隐藏后取消注册，不需要再次触发buttonPress信号
    m_regionInter->unregisterRegion();
}

void DockPopupWindow::compositeChanged()
{
    if (m_wmHelper->hasComposite())
        setBorderColor(QColor(255, 255, 255, 255 * 0.05));
    else
        setBorderColor(QColor("#2C3238"));
}

void DockPopupWindow::ensureRaised()
{
    if (isVisible()) {
        QWidget *content = getContent();
        if (!content || !content->isVisible()) {
            this->setVisible(false);
        } else {
            raise();
            if (!windowFlags().testFlag(Qt::WindowDoesNotAcceptFocus))
                activateWindow();
        }
    }
}

/**
 * @brief 将缩放后的点的位置按照缩放比进行放大
 *  之所以这么复杂，是因为qt在多屏缩放的情况下，非首屏（即topLeft为(0,0)的屏幕）上的点坐标错误
 *  详见QTBUG-81695
 * @param point
 * @return QPoint
 */
QPoint DockPopupWindow::scalePoint(QPoint point)
{
    const qreal pixelRatio = qApp->devicePixelRatio();
    QScreen * const screen = qApp->screenAt(point);
    if (!screen) {
        qWarning() << "Cannot find screen the point is located: " << point;
        return point;
    }

    const QRect &screenRect = screen->geometry();

    qreal pointX = point.x() * pixelRatio;
    if (screenRect.x() != 0) {
        pointX = (screenRect.x() / pixelRatio + point.x() - screenRect.x()) * pixelRatio;
    }

    qreal pointY = point.y() * pixelRatio;
    if (screenRect.y() != 0) {
        pointY = (screenRect.y() / pixelRatio + point.y() - screenRect.y()) * pixelRatio;
    }

    return QPoint(pointX, pointY);
}
