// SPDX-FileCopyrightText: 2019 - 2022 UnionTech Software Technology Co., Ltd.
//
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "mainpanelcontrol.h"
#include "dockitem.h"
#include "overflowitem.h"
#include "placeholderitem.h"
#include "components/appdrag.h"
#include "appitem.h"
#include "pluginsitem.h"
#include "traypluginitem.h"
#include "dockitemmanager.h"
#include "touchsignalmanager.h"
#include "utils.h"
#include "desktop_widget.h"

#include <QDrag>
#include <QTimer>
#include <QStandardPaths>
#include <QString>
#include <QApplication>
#include <QPointer>
#include <QBoxLayout>
#include <QLabel>
#include <QPixmap>
#include <QBoxLayout>
#include <QtConcurrent/QtConcurrentRun>
#include <qpa/qplatformnativeinterface.h>
#include <qpa/qplatformintegration.h>
#define protected public
#include <QtGui/private/qsimpledrag_p.h>
#undef protected
#include <QtGui/private/qshapedpixmapdndwindow_p.h>
#include <QtGui/private/qguiapplication_p.h>

#include <DGuiApplicationHelper>
#include <DWindowManagerHelper>

#define SPLITER_SIZE 2
#define TRASH_MARGIN 20
#define PLUGIN_MAX_SIZE  40
#define PLUGIN_MIN_SIZE  20
#define DESKTOP_SIZE  10

DWIDGET_USE_NAMESPACE

const QString APPOVERFLOWAREA_NAME = "appoverflowarea";

// record the message to used in resize
struct ResizeMessage
{
    int realAppItemSize;
    int maxCount;
    bool showOverflow;
};

MainPanelControl::MainPanelControl(QWidget *parent)
    : QWidget(parent)
    , m_maxAppStayCount(-1)
    , m_showOverflow(false)
    , m_appItemSize(0)
    , m_isDraging(false)
    , m_mainPanelLayout(new QBoxLayout(QBoxLayout::LeftToRight, this))
    , m_fixedAreaWidget(new QWidget(this))
    , m_fixedAreaLayout(new QBoxLayout(QBoxLayout::LeftToRight))
    , m_fixedSpliter(new QLabel(this))
    , m_appAreaWidget(new QWidget(this))
    , m_appAreaSonWidget(new QWidget(this))
    , m_appAreaSonLayout(new QBoxLayout(QBoxLayout::LeftToRight))
    , m_appSpliter(new QLabel(this))
    , m_appOverflowWidget(new QWidget(this))
    , m_appOverflowLayout(new QBoxLayout(QBoxLayout::LeftToRight))
    , m_overflowItem(new OverflowItem(this))
    , m_overflowRecentlyUsedApp(nullptr)
    , m_overflowSpliter(new QLabel(this))
    , m_overflowAppInter(new DBusDock("com.deepin.dde.daemon.Dock", "/com/deepin/dde/daemon/Dock", QDBusConnection::sessionBus(), this))
    , m_overflowPopMargin(10)
    , m_trayAreaWidget(new QWidget(this))
    , m_trayAreaLayout(new QBoxLayout(QBoxLayout::LeftToRight))
    , m_traySpliter(new QLabel(this))
    , m_pluginAreaWidget(new QWidget(this))
    , m_pluginLayout(new QBoxLayout(QBoxLayout::LeftToRight))
    , m_desktopWidget(new DesktopWidget(this))
    , m_position(Position::Bottom)
    , m_placeholderItem(nullptr)
    , m_appDragWidget(nullptr)
    , m_dislayMode(Efficient)
    , m_tray(nullptr)
    , m_dragIndex(-1)
    , m_trashItem(nullptr)
    , m_updateDelay(new QTimer(this))
{
    initUI();
    updateMainPanelLayout();
    setAcceptDrops(true);
    setMouseTracking(true);
    m_updateDelay->setInterval(0);
    m_updateDelay->setSingleShot(true);

    m_appAreaWidget->installEventFilter(this);
    m_appAreaSonWidget->installEventFilter(this);
    m_appOverflowWidget->installEventFilter(this);
    m_trayAreaWidget->installEventFilter(this);
    m_pluginAreaWidget->installEventFilter(this);

    // 在设置每条线大小前，应该设置fixedsize(0,0)
    // 应为paintEvent函数会先调用设置背景颜色，大小为随机值
    m_fixedSpliter->setFixedSize(0, 0);
    m_overflowSpliter->setFixedSize(0, 0);
    m_appSpliter->setFixedSize(0, 0);
    m_traySpliter->setFixedSize(0, 0);

    connect(m_updateDelay, &QTimer::timeout, this, [=] {
        qDebug() << "Layout is refreshed";
        updatePluginsLayout();
        resizeDockIcon();
        m_overflowItem->hidePopUpWindow();
        m_overflowItem->refreshIcon();
        if (m_showOverflow) {
            emit requestUpdateLayout(false);
        }
    });
    // when size or new item in , resizeLayout
    connect(this, &MainPanelControl::requestUpdateLayout, this, [ this ] (bool needAdd, bool forceShow) {
        m_overflowItem->hidePopUpWindow();
        updateLayout(needAdd, forceShow);
    });

    connect(DockItemManager::instance(), &DockItemManager::requestUpdateActiveApp, this, [ this ] (int index) {
        // 不属于溢出区app
        if ( (!m_showOverflow) || (index <= m_maxAppStayCount) ) {
            return;
        }

        const auto children = DockItemManager::instance()->itemList();
        updateOverflowRecentlyUsedApp(qobject_cast<AppItem *>(children[index].data()));

        // 更新溢出区打开记录
        if (m_overflowAppOpenRecord.size() > 30) {
            for (int i = 0; i < 10; i++) {
                m_overflowAppOpenRecord.removeFirst();
            }
        }
        for (int i = 0; i < m_overflowAppOpenRecord.size(); ++i) {
            // 如果存在该应用记录，先将原记录删除
            if (m_overflowAppOpenRecord.at(i).first == qobject_cast<AppItem *>(children[index].data())->appId()) {
                m_overflowAppOpenRecord.removeAt(i);
                break;
            }
        }
        m_overflowAppOpenRecord.append(QPair<QString, int>(qobject_cast<AppItem *>(children[index].data())->appId(), index));
    });

    connect(DockItemManager::instance(), &DockItemManager::hideOverflowPopup, this, [ this ] {
        if (!m_showOverflow) {
            return;
        }
        m_overflowItem->hidePopUpWindow();
    });

    connect(this, &MainPanelControl::maskAlphaChanged, m_overflowItem, &OverflowItem::maskAlphaChanged, Qt::QueuedConnection);

    connect(DockItemManager::instance(), &DockItemManager::requestAppItemCloseWindow, this, [ this ] {
        if (!m_showOverflow) {
            return;
        }
        emit requestUpdateLayout(true);
    });
    connect(m_overflowItem, &OverflowItem::popupWindowCursorOut, this, &MainPanelControl::overflowPopupCursorOut);
    connect(m_overflowItem, &OverflowItem::popupWindowShowChanged, this,  &MainPanelControl::overflowPopupShowChanged);
}

MainPanelControl::~MainPanelControl()
{
    if (m_overflowRecentlyUsedApp) {
        delete m_overflowRecentlyUsedApp;
        m_overflowRecentlyUsedApp = nullptr;
    }
}

void MainPanelControl::initUI()
{
    /* 固定区域 */
    m_fixedAreaWidget->setObjectName("fixedarea");
    m_fixedAreaWidget->setLayout(m_fixedAreaLayout);
    m_fixedAreaLayout->setSpacing(0);
    m_fixedAreaLayout->setContentsMargins(0, 0, 0, 0);
    m_mainPanelLayout->addWidget(m_fixedAreaWidget);

    m_fixedSpliter->setObjectName("spliter_fix");
    m_mainPanelLayout->addWidget(m_fixedSpliter);

    /* 应用程序区域 */
    m_appAreaWidget->setAccessibleName("AppFullArea");
    m_mainPanelLayout->addWidget(m_appAreaWidget);
    m_appAreaSonWidget->setObjectName("apparea");
    m_appAreaSonWidget->setLayout(m_appAreaSonLayout);
    m_appAreaSonLayout->setSpacing(0);
    m_appAreaSonLayout->setContentsMargins(0, 0, 0, 0);

    m_appSpliter->setObjectName("spliter_app");
    m_mainPanelLayout->addWidget(m_appSpliter);

    /* overflow zone */
    m_appOverflowWidget->setAccessibleName("AppOverFlowArea");
    m_appOverflowWidget->setObjectName(APPOVERFLOWAREA_NAME);
    m_mainPanelLayout->addWidget(m_appOverflowWidget);
    m_appOverflowWidget->setLayout(m_appOverflowLayout);
    m_overflowItem->setVisible(false);
    m_appOverflowLayout->addWidget(m_overflowItem);
    m_appOverflowLayout->setSpacing(0);
    m_appOverflowLayout->setContentsMargins(0, 0, 0, 0);

    m_overflowSpliter->setObjectName("spliter_overflow");
    m_mainPanelLayout->addWidget(m_overflowSpliter);

    /* 托盘区域 */
    m_trayAreaWidget->setObjectName("trayarea");
    m_trayAreaWidget->setLayout(m_trayAreaLayout);
    m_trayAreaLayout->setSpacing(0);
    m_trayAreaLayout->setContentsMargins(0, 10, 0, 10);
    m_mainPanelLayout->addWidget(m_trayAreaWidget);

    m_traySpliter->setObjectName("spliter_tray");
    m_mainPanelLayout->addWidget(m_traySpliter);

    /* 插件区域 */
    m_pluginAreaWidget->setObjectName("pluginarea");
    m_pluginAreaWidget->setLayout(m_pluginLayout);
    m_pluginLayout->setSpacing(10);
    m_pluginLayout->setContentsMargins(0, 0, 0, 0);
    m_mainPanelLayout->addWidget(m_pluginAreaWidget, 0, Qt::AlignCenter);

    /* 桌面预览 */
    m_desktopWidget->setObjectName("showdesktoparea");
    m_mainPanelLayout->addWidget(m_desktopWidget);

    m_mainPanelLayout->setSpacing(0);
    m_mainPanelLayout->setContentsMargins(0, 0, 0, 0);
    m_mainPanelLayout->setAlignment(m_fixedSpliter, Qt::AlignCenter);
    m_mainPanelLayout->setAlignment(m_appSpliter, Qt::AlignCenter);
    m_mainPanelLayout->setAlignment(m_overflowSpliter, Qt::AlignCenter);
    m_mainPanelLayout->setAlignment(m_traySpliter, Qt::AlignCenter);
}

/**
 * @brief MainPanelControl::setDisplayMode 根据任务栏显示模式更新界面显示，如果是时尚模式，没有‘显示桌面'区域，否则就有
 * @param dislayMode 任务栏显示模式
 */
void MainPanelControl::setDisplayMode(DisplayMode dislayMode)
{
    if (dislayMode == m_dislayMode)
        return;
    m_dislayMode = dislayMode;
    updateDisplayMode();
}

/**根据任务栏在屏幕上的位置，更新任务栏各控件布局
 * @brief MainPanelControl::updateMainPanelLayout
 */
void MainPanelControl::updateMainPanelLayout()
{
    m_overflowItem->setLayoutPosition(m_position);
    switch (m_position) {
    case Position::Top:
    case Position::Bottom:
        m_fixedAreaWidget->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding);
        m_appOverflowWidget->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding);
        m_appAreaWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
        m_pluginAreaWidget->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
        m_trayAreaWidget->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Expanding);
        m_mainPanelLayout->setDirection(QBoxLayout::LeftToRight);
        m_fixedAreaLayout->setDirection(QBoxLayout::LeftToRight);
        m_appOverflowLayout->setDirection(QBoxLayout::LeftToRight);
        m_pluginLayout->setDirection(QBoxLayout::LeftToRight);
        m_trayAreaLayout->setDirection(QBoxLayout::LeftToRight);
        m_appAreaSonLayout->setDirection(QBoxLayout::LeftToRight);
        m_trayAreaLayout->setContentsMargins(0, 10, 0, 10);
        m_pluginLayout->setContentsMargins(10, 0, 10, 0);
        break;
    case Position::Right:
    case Position::Left:
        m_fixedAreaWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
        m_appOverflowWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
        m_appAreaWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
        m_pluginAreaWidget->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
        m_trayAreaWidget->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
        m_mainPanelLayout->setDirection(QBoxLayout::TopToBottom);
        m_fixedAreaLayout->setDirection(QBoxLayout::TopToBottom);
        m_appOverflowLayout->setDirection(QBoxLayout::TopToBottom);
        m_pluginLayout->setDirection(QBoxLayout::TopToBottom);
        m_trayAreaLayout->setDirection(QBoxLayout::TopToBottom);
        m_appAreaSonLayout->setDirection(QBoxLayout::TopToBottom);
        m_trayAreaLayout->setContentsMargins(10, 0, 10, 0);
        m_pluginLayout->setContentsMargins(0, 10, 0, 10);
        break;
    }

    // 显示桌面的区域
    resizeDesktopWidget();

    // 设置任务栏各区域图标大小
    m_updateDelay->start();
}

/**往固定区域添加应用
 * @brief MainPanelControl::addFixedAreaItem
 * @param index　位置索引，如果为负数则插入到最后，为正则插入到指定位置
 * @param wdg　应用指针对象
 */
void MainPanelControl::addFixedAreaItem(int index, QWidget *wdg)
{
    if(m_position == Position::Top || m_position == Position::Bottom){
        wdg->setMaximumSize(height(),height());
    } else {
        wdg->setMaximumSize(width(),width());
    }
    m_fixedAreaLayout->insertWidget(index, wdg);
}

/**往应用区域添加应用
 * @brief MainPanelControl::addAppAreaItem
 * @param index　位置索引，如果为负数则插入到最后，为正则插入到指定位置
 * @param wdg　应用指针对象
 */
void MainPanelControl::addAppAreaItem(int index, QWidget *wdg)
{
    if(m_position == Position::Top || m_position == Position::Bottom){
        wdg->setMaximumSize(height(),height());
    } else {
        wdg->setMaximumSize(width(),width());
    }
    m_appAreaSonLayout->insertWidget(index, wdg);
}

/**往托盘插件区域添加应用
 * @brief MainPanelControl::addTrayAreaItem
 * @param index　位置索引，如果为负数则插入到最后，为正则插入到指定位置
 * @param wdg　应用指针对象
 */
void MainPanelControl::addTrayAreaItem(int index, QWidget *wdg)
{
    m_tray = static_cast<TrayPluginItem *>(wdg);
    m_trayAreaLayout->insertWidget(index, wdg);
}

/**往插件区域添加应用，保存回收站插件指针对象
 * @brief MainPanelControl::addPluginAreaItem
 * @param index　位置索引，如果为负数则插入到最后，为正则插入到指定位置
 * @param wdg　应用指针对象
 */
void MainPanelControl::addPluginAreaItem(int index, QWidget *wdg)
{
    //因为日期时间插件和其他插件的大小有异，为了方便设置边距，在插件区域布局再添加一层布局设置边距
    //因此在处理插件图标时，需要通过两层布局判断是否为需要的插件，例如拖动插件位置等判断
    QBoxLayout * boxLayout = new QBoxLayout(QBoxLayout::LeftToRight);
    boxLayout->addWidget(wdg, 0, Qt::AlignCenter);
    m_pluginLayout->insertLayout(index, boxLayout, 0);

    // 保存垃圾箱插件指针
    PluginsItem *pluginsItem = qobject_cast<PluginsItem *>(wdg);
    if (pluginsItem && pluginsItem->pluginName() == "trash")
        m_trashItem = pluginsItem;
}

/**移除固定区域某一应用
 * @brief MainPanelControl::removeFixedAreaItem
 * @param wdg 应用指针对象
 */
void MainPanelControl::removeFixedAreaItem(QWidget *wdg)
{
    m_fixedAreaLayout->removeWidget(wdg);
}

/**移除应用区域某一应用
 * @brief MainPanelControl::removeAppAreaItem
 * @param wdg 应用指针对象
 */
void MainPanelControl::removeAppAreaItem(QWidget *wdg)
{
    m_appAreaSonLayout->removeWidget(wdg);
}

/**移除托盘插件区域某一应用
 * @brief MainPanelControl::removeTrayAreaItem
 * @param wdg 应用指针对象
 */
void MainPanelControl::removeTrayAreaItem(QWidget *wdg)
{
    m_trayAreaLayout->removeWidget(wdg);
}

/**移除插件区域某一应用
 * @brief MainPanelControl::removePluginAreaItem
 * @param wdg 应用指针对象
 */
void MainPanelControl::removePluginAreaItem(QWidget *wdg)
{
    //因为日期时间插件大小和其他插件有异，为了方便设置边距，各插件中增加一层布局
    //因此remove插件图标时，需要从多的一层布局中取widget进行判断是否需要移除的插件
    // 清空保存的垃圾箱插件指针
    PluginsItem *pluginsItem = qobject_cast<PluginsItem *>(wdg);
    if (pluginsItem && pluginsItem->pluginName() == "trash")
        m_trashItem = nullptr;

    for (int i = 0; i < m_pluginLayout->count(); ++i) {
        QLayoutItem *layoutItem = m_pluginLayout->itemAt(i);
        QLayout *boxLayout = layoutItem->layout();
        if (boxLayout && boxLayout->itemAt(0)->widget() == wdg) {
            boxLayout->removeWidget(wdg);
            m_pluginLayout->removeItem(layoutItem);
            delete layoutItem;
            layoutItem = nullptr;
        }
    }
}

void MainPanelControl::resizeEvent(QResizeEvent *event)
{
    //先通过消息循环让各部件调整好size后再计算图标大小
    //避免因为部件size没有调整完导致计算的图标大小不准确
    //然后重复触发m_pluginAreaWidget的reszie事件并重复计算，造成任务栏图标抖动问题
    QWidget::resizeEvent(event);
    resizeDesktopWidget();
    resizeDockIcon();
}

/**根据任务栏所在位置， 设置应用区域控件的大小
 * @brief MainPanelControl::updateAppAreaSonWidgetSize
 */
void MainPanelControl::updateAppAreaSonWidgetSize()
{
    if ((m_position == Position::Top) || (m_position == Position::Bottom)) {
        m_appAreaSonWidget->setMaximumHeight(this->height());
        m_appAreaSonWidget->setMaximumWidth(m_appAreaWidget->width());
    } else {
        m_appAreaSonWidget->setMaximumWidth(this->width());
        m_appAreaSonWidget->setMaximumHeight(m_appAreaWidget->height());
    }

    m_appAreaSonWidget->adjustSize();

    moveAppSonWidget();
}

/**
 * @brief setPositonValue 根据传入的位置更新界面布局，比如任务栏在左，布局应该是上下布局，任务栏在下，应该是左右布局
 * @param position 任务栏的位置
 */
void MainPanelControl::setPositonValue(Dock::Position position)
{
    if (m_position == position)
        return;

    m_position = position;

    QTimer::singleShot(0, this, &MainPanelControl::updateMainPanelLayout);
}

/**向任务栏插入各类应用,并将属于同一个应用的窗口合并到同一个应用图标
 * @brief MainPanelControl::insertItem
 * @param index 位置索引
 * @param item 应用指针对象
 */
void MainPanelControl::insertItem(int index, DockItem *item)
{
    if (!item)
        return;

    item->installEventFilter(this);

    switch (item->itemType()) {
    case DockItem::Launcher:
        addFixedAreaItem(0, item);
        break;
    case DockItem::FixedPlugin:
        addFixedAreaItem(index, item);
        break;
    case DockItem::App:
    case DockItem::Placeholder:
        item->setVisible(true);
        addAppAreaItem(index, item);
        break;
    case DockItem::TrayPlugin: // 此处只会有一个tray系统托盘插件，微信、声音、网络蓝牙等等，都在系统托盘插件中处理的
        addTrayAreaItem(index, item);
        break;
    case DockItem::Plugins:
        addPluginAreaItem(index, item);
        break;
    default:
        break;
    }

    item->checkEntry();

    m_updateDelay->start();
}

/**从任务栏移除某一应用，并更新任务栏图标大小
 * @brief MainPanelControl::removeItem
 * @param item 应用指针对象
 */
void MainPanelControl::removeItem(DockItem *item)
{
    switch (item->itemType()) {
    case DockItem::Launcher:
    case DockItem::FixedPlugin:
        removeFixedAreaItem(item);
        break;
    case DockItem::App:
    case DockItem::Placeholder:
        item->setVisible(false);
        removeAppAreaItem(item);
        break;
    case DockItem::TrayPlugin:
        removeTrayAreaItem(item);
        break;
    case DockItem::Plugins:
        removePluginAreaItem(item);
        break;
    default:
        break;
    }

    item->removeEventFilter(this);

    /** 此处重新计算大小的时候icon的个数在原有个数上减少了一个，导致每个icon的大小跟原来大小不一致，需要重新设置setFixedSize
     *  在龙芯处理器上当app数量过多时，会导致拖动app耗时严重，造成卡顿
     *  注意:不能屏蔽此接口，否则会造成插件移除时无法更新icon大小
     */
    m_updateDelay->start();
}

/**任务栏移动应用图标
 * @brief MainPanelControl::moveItem
 * @param sourceItem 即将插入的应用
 * @param targetItem 被移动的应用
 */
void MainPanelControl::moveItem(DockItem *sourceItem, DockItem *targetItem)
{
    // get target index
    int idx = getItemIndex(targetItem);
    if (-1 == idx)
        return;

    // remove old item
    removeItem(sourceItem);

    // insert new position
    if (sourceItem->isDragging()) {
        m_dragIndex = idx;
    }
    insertItem(idx, sourceItem);
}

int MainPanelControl::getItemIndex(DockItem *targetItem) const
{
    if (!targetItem)
        return -1;

    if (targetItem->itemType() == DockItem::App)
        return m_appAreaSonLayout->indexOf(targetItem);

    if (targetItem->itemType() == DockItem::Plugins) {
        //因为日期时间插件大小和其他插件大小有异，为了设置边距，在各插件中增加了一层布局
        //因此有拖动图标时，需要从多的一层布局中判断是否相同插件而获取插件位置顺序
        for (int i = 0; i < m_pluginLayout->count(); ++i) {
            QLayout *layout = m_pluginLayout->itemAt(i)->layout();
            if (layout && layout->itemAt(0)->widget() == targetItem) {
                return i;
            }
        }
    }

    if (targetItem->itemType() == DockItem::FixedPlugin)
        return m_fixedAreaLayout->indexOf(targetItem);

    return -1;
}

void MainPanelControl::dragEnterEvent(QDragEnterEvent *e)
{
    //拖拽图标到任务栏时，如果拖拽到垃圾箱插件图标widget上，则默认不允许拖拽，其他位置默认为允许拖拽
    QWidget *widget = QApplication::widgetAt(QCursor::pos());
    //"trash-centralwidget"名称是在PluginsItem类中m_centralWidget->setObjectName(pluginInter->pluginName() + "-centralwidget");
    if (widget && widget->objectName() == "trash-centralwidget") {
        return;
    }

    e->accept();
}

void MainPanelControl::dragLeaveEvent(QDragLeaveEvent *e)
{
    Q_UNUSED(e);
    if (m_placeholderItem) {
        const QRect r(static_cast<QWidget *>(parent())->pos(), size());
        const QPoint p(QCursor::pos());

        // remove margins to fix a touch screen bug:
        // the mouse point position will stay on this rect's margins after
        // drag move to the edge of screen
        if (r.marginsRemoved(QMargins(1, 10, 1, 1)).contains(p))
            return;

        removeAppAreaItem(m_placeholderItem);
        m_placeholderItem->deleteLater();
        m_isDraging = false;
    }
}

void MainPanelControl::dropEvent(QDropEvent *e)
{
    if (m_placeholderItem) {
        emit itemAdded(e->mimeData()->data(m_draggingMimeKey), m_appAreaSonLayout->indexOf(m_placeholderItem));

        removeAppAreaItem(m_placeholderItem);
        m_placeholderItem->deleteLater();
        m_isDraging = false;
    }
}

void MainPanelControl::handleDragMove(QDragMoveEvent *e, bool isFilter)
{
    if (!e->source()) {
        // 应用程序拖到dock上
        e->accept();

        m_isDraging = true;

        DockItem *insertPositionItem = dropTargetItem(nullptr, e->pos());

        if (m_placeholderItem.isNull()) {
            m_placeholderItem = new PlaceholderItem;

            if (m_position == Dock::Top || m_position == Dock::Bottom) {
                if (m_appAreaSonWidget->mapFromParent(e->pos()).x() > m_appAreaSonWidget->rect().right()) {
                    // 插入到最右侧
                    insertPositionItem = nullptr;
                }
            } else {
                if (m_appAreaSonWidget->mapFromParent(e->pos()).y() > m_appAreaSonWidget->rect().bottom()) {
                    // 插入到最下测
                    insertPositionItem = nullptr;
                }
            }

            insertItem(m_appAreaSonLayout->indexOf(insertPositionItem), m_placeholderItem);

        } else if (insertPositionItem && m_placeholderItem != insertPositionItem) {
            moveItem(m_placeholderItem, insertPositionItem);
        }

        return;
    }

    DockItem *sourceItem = qobject_cast<DockItem *>(e->source());

    if (!sourceItem) {
        e->ignore();
        return;
    }

    DockItem *targetItem = nullptr;

    if (isFilter) {
        // appItem调整顺序或者移除驻留
        targetItem = dropTargetItem(sourceItem, mapFromGlobal(m_appDragWidget->mapToGlobal(e->pos())));

        if (targetItem) {
            m_appDragWidget->setOriginPos((m_appAreaSonWidget->mapToGlobal(targetItem->pos())));
        } else {
            targetItem = sourceItem;
        }
    } else {
        // other dockItem调整顺序
        targetItem = dropTargetItem(sourceItem, e->pos());
    }

    if (!targetItem) {
        e->ignore();
        return;
    }

    e->accept();

    if (targetItem == sourceItem)
        return;

    moveItem(sourceItem, targetItem);
    emit itemMoved(sourceItem, targetItem);
}

void MainPanelControl::dragMoveEvent(QDragMoveEvent *e)
{
    DockItem *sourceItem = qobject_cast<DockItem *>(e->source());
    if (sourceItem) {
        // 溢出区app不处理
        if (sourceItem->itemType() == DockItem::App && !m_appDragWidget->canDrag()) {
            return;
        }
        handleDragMove(e, false);
        return;
    }

    // 拖app到dock上
    const char *RequestDockKey = "RequestDock";
    const char *RequestDockKeyFallback = "text/plain";
    const char *DesktopMimeType = "application/x-desktop";
    auto DragmineData = e->mimeData();

    m_draggingMimeKey = DragmineData->formats().contains(RequestDockKey) ? RequestDockKey : RequestDockKeyFallback;

    // dragging item is NOT a desktop file
    if (QMimeDatabase().mimeTypeForFile(DragmineData->data(m_draggingMimeKey)).name() != DesktopMimeType) {
        m_draggingMimeKey.clear();
        e->setAccepted(false);
        qDebug() << "dragging item is NOT a desktop file";
        return;
    }

    //如果当前从桌面拖拽的的app是trash，则不能放入app任务栏中
    QString str = "file://";
    //启动器
    QString str_t = "";

    str.append(QStandardPaths::locate(QStandardPaths::DesktopLocation, "dde-trash.desktop"));
    str_t.append(QStandardPaths::locate(QStandardPaths::ApplicationsLocation, "dde-trash.desktop"));

    if ((str == DragmineData->data(m_draggingMimeKey)) || (str_t == DragmineData->data(m_draggingMimeKey))) {
        e->setAccepted(false);
        return;
    }

    if (appIsOnDock(DragmineData->data(m_draggingMimeKey))) {
        e->setAccepted(false);
        return;
    }

    handleDragMove(e, false);
}

bool MainPanelControl::eventFilter(QObject *watched, QEvent *event)
{
    // 更新应用区域大小和任务栏图标大小
    if (watched == m_appAreaSonWidget) {
        switch (event->type()) {
        case QEvent::LayoutRequest:
            m_appAreaSonWidget->adjustSize();
            resizeDockIcon();
            break;
        case QEvent::Resize:
            resizeDockIcon();
            break;
        default:
            moveAppSonWidget();
            break;
        }
    }

    // fix:88133 在计算icon大小时m_pluginAreaWidget的数据错误
    if (watched == m_pluginAreaWidget) {
        switch (event->type()) {
        case QEvent::Resize:
            resizeDockIcon();
            break;
        default:
            break;
        }
    }

    // 更新应用区域子控件大小以及位置
    if (watched == m_appAreaWidget) {
        if (event->type() == QEvent::Resize)
            updateAppAreaSonWidgetSize();

        if (event->type() == QEvent::Move)
            moveAppSonWidget();
    }

    if (m_appDragWidget && watched == static_cast<QGraphicsView *>(m_appDragWidget)->viewport()) {
        bool isContains = rect().contains(mapFromGlobal(QCursor::pos()));
        if (isContains && m_appDragWidget->canDrag()) {
            if (event->type() == QEvent::DragMove) {
                handleDragMove(static_cast<QDragMoveEvent *>(event), true);
            } else if (event->type() == QEvent::Drop) {
                m_appDragWidget->hide();
                return true;
            }
        }

        return false;
    }

    if (event->type() != QEvent::MouseMove)
        return false;

    QMouseEvent *mouseEvent = dynamic_cast<QMouseEvent *>(event);
    if (!mouseEvent || mouseEvent->buttons() != Qt::LeftButton)
        return false;

    DockItem *item = qobject_cast<DockItem *>(watched);
    if (!item)
        return false;

    if (item->itemType() != DockItem::App && item->itemType() != DockItem::Plugins && item->itemType() != DockItem::FixedPlugin)
        return false;

    const QPoint pos = mouseEvent->globalPos();
    const QPoint distance = pos - m_mousePressPos;
    if (distance.manhattanLength() < QApplication::startDragDistance())
        return false;

    // source为MouseEventSynthesizedByQt时，事件由触屏事件转换而来，触屏没有收到后端的延迟触屏信号时不进行拖动
    if (mouseEvent->source() == Qt::MouseEventSynthesizedByQt && !TouchSignalManager::instance()->isDragIconPress()) {
        return false;
    }

    static const QGSettings *g_settings = Utils::ModuleSettingsPtr("app");
    if (!g_settings || !g_settings->keys().contains("removeable") || g_settings->get("removeable").toBool())
        startDrag(item);

    return QWidget::eventFilter(watched, event);
}

void MainPanelControl::enterEvent(QEvent *event)
{
    if (Utils::IS_WAYLAND_DISPLAY) {
        Utils::updateCursor(this);
    }

    QWidget::enterEvent(event);
}

void MainPanelControl::mousePressEvent(QMouseEvent *e)
{
    if (e->button() == Qt::LeftButton) {
        m_mousePressPos = e->globalPos();
    }

    QWidget::mousePressEvent(e);
}

void MainPanelControl::startDrag(DockItem *dockItem)
{
    QPointer<DockItem> item = dockItem;
    const QPixmap pixmap = item->grab();

    item->setDragging(true);
    m_isDraging = true;
    item->update();

    auto itemParent = item->parent();
    bool isOverflowItem = (itemParent && (itemParent->objectName() == OVERFLOWWIDGET_OBJECT_NAME || itemParent->objectName() == APPOVERFLOWAREA_NAME));
    QDrag *drag = nullptr;
    // INFO: update the value of drag
    auto releaseDrag = [this](bool forceShow = true) {
        if (!m_isDraging) {
            return;
        }
        m_isDraging = false;
        emit requestUpdateLayout(false, forceShow);
    };
    if (item->itemType() == DockItem::App) {
        AppDrag *appDrag = new AppDrag(item);

        m_appDragWidget = appDrag->appDragWidget();
        m_appDragWidget->setCanDrag(!isOverflowItem);

        if (Utils::IS_WAYLAND_DISPLAY) {
            m_dragIndex = getItemIndex(dockItem);
            connect(m_appDragWidget, &AppDragWidget::requestRemoveSelf, this, [ = ](bool needDelete) {
                m_appDragWidget = nullptr;
                AppItem *appItem = qobject_cast<AppItem *>(item);
                if (item.isNull() || !appItem->isValid())
                    return;

                bool canInsertItem = !AppDragWidget::isRemoveAble() || appItem->hasWindows();
                if (m_dragIndex > -1 && !needDelete) {
                    if (canInsertItem)
                        insertItem(m_dragIndex, item);
                    m_dragIndex = -1;
                } else {
                    if (-1 == m_appAreaSonLayout->indexOf(item) && m_dragIndex != -1) {
                        if (canInsertItem)
                            insertItem(m_dragIndex, item);
                        m_dragIndex = -1;
                    }

                    // 溢出区域内的应用不能移除
                    if (!isOverflowItem) {
                        AppItem *app = qobject_cast<AppItem *>(item.data());
                        app->undock();
                    }
                }
                item->setDragging(false);
                item->update();
                releaseDrag(canInsertItem);
            });
        } else {
            connect(m_appDragWidget, &AppDragWidget::destroyed, this, [ = ] {
                m_appDragWidget = nullptr;
                AppItem *appItem = qobject_cast<AppItem *>(item);
                if (item.isNull() || !appItem->isValid())
                    return ;

                bool canInsertItem = !AppDragWidget::isRemoveAble() || appItem->hasWindows();
                if (-1 == m_appAreaSonLayout->indexOf(item) && m_dragIndex != -1) {
                    if (canInsertItem)
                        insertItem(m_dragIndex, item);
                    m_dragIndex = -1;
                }
                item->setDragging(false);
                item->update();
                releaseDrag(canInsertItem);
            });
        }

        connect(m_appDragWidget, &AppDragWidget::requestRemoveItem, this, [ = ] {
            if (-1 != m_appAreaSonLayout->indexOf(item)) {
                m_dragIndex = m_appAreaSonLayout->indexOf(item);
                removeItem(item);
            }
        });

        appDrag->appDragWidget()->setOriginPos((m_appAreaSonWidget->mapToGlobal(item->pos())));
        appDrag->appDragWidget()->setDockInfo(m_position, QRect(mapToGlobal(pos()), size()));
        const QPixmap &dragPix = qobject_cast<AppItem *>(item)->appIcon();

        appDrag->setPixmap(dragPix);
        m_appDragWidget->show();

        if (DWindowManagerHelper::instance()->hasComposite()) {
            static_cast<QGraphicsView *>(m_appDragWidget)->viewport()->installEventFilter(this);
        } else {
            appDrag->QDrag::setPixmap(dragPix);
        }

        drag = appDrag;
        drag->setHotSpot(dragPix.rect().center() / dragPix.devicePixelRatioF());
    } else {
        drag = new QDrag(item);
        drag->setPixmap(pixmap);
        drag->setHotSpot(pixmap.rect().center() / pixmap.devicePixelRatioF());
    }

    // isNeedBack 保存是否需要重置垃圾箱的AcceptDrops
    // 设置垃圾箱插件AcceptDrops false
    bool isNeedBack = false;
    if (item->itemType() == DockItem::Plugins && m_trashItem && dockItem != m_trashItem) {
        m_trashItem->centralWidget()->setAcceptDrops(false);
        isNeedBack = true;
    }

    auto mimeData = new QMimeData;
    QByteArray data;
    QDataStream dataStream(&data, QIODevice::ReadWrite);
    dataStream << pixmap;
    mimeData->setData("application/x-desktop", data);
    drag->setMimeData(mimeData);

    drag->exec(Qt::MoveAction);

    if (item->itemType() != DockItem::App || m_dragIndex == -1) {
        m_appDragWidget = nullptr;
        item->setDragging(false);
        item->update();
        releaseDrag();

        // isNeedBack是否需要设置垃圾箱插件AcceptDrops true
        if (isNeedBack)
            m_trashItem->centralWidget()->setAcceptDrops(true);
    }
}

DockItem *MainPanelControl::dropTargetItem(DockItem *sourceItem, QPoint point)
{
    QWidget *parentWidget = m_appAreaSonWidget;

    if (sourceItem) {
        switch (sourceItem->itemType()) {
        case DockItem::App:
            parentWidget = m_appAreaSonWidget;
            break;
        case DockItem::Plugins:
            parentWidget = m_pluginAreaWidget;
            break;
        case DockItem::FixedPlugin:
            parentWidget = m_fixedAreaWidget;
            break;
        default:
            break;
        }
    }

    if (!parentWidget)
        return nullptr;

    point = parentWidget->mapFromParent(point);
    QLayout *parentLayout = parentWidget->layout();

    DockItem *targetItem = nullptr;

    for (int i = 0 ; i < parentLayout->count(); ++i) {
        QLayoutItem *layoutItem = parentLayout->itemAt(i);

        DockItem *dockItem = nullptr;
        if (parentWidget == m_pluginAreaWidget) {
            QLayout *layout = layoutItem->layout();
            if (layout) {
                dockItem = qobject_cast<DockItem *>(layout->itemAt(0)->widget());
            }
        } else{
            dockItem = qobject_cast<DockItem *>(layoutItem->widget());
        }

        if (!dockItem)
            continue;

        QRect rect(dockItem->pos(), dockItem->size());
        if (rect.contains(point)) {
            targetItem = dockItem;
            break;
        }
    }

    if (!targetItem && parentWidget == m_appAreaSonWidget) {
        // appitem调整顺序是，判断是否拖放在两边空白区域
        targetItem = sourceItem;
    }

    return targetItem;
}

void MainPanelControl::updateDisplayMode()
{
    moveAppSonWidget();
    resizeDesktopWidget();
}

/**把驻留应用和被打开的应用所在窗口移动到指定位置
 * @brief MainPanelControl::moveAppSonWidget
 */
void MainPanelControl::moveAppSonWidget()
{
    QRect rect(QPoint(0, 0), m_appAreaSonWidget->size());
    if (DisplayMode::Efficient == m_dislayMode) {
        switch (m_position) {
        case Top:
        case Bottom :
            rect.moveTo(m_appAreaWidget->pos());
            break;
        case Right:
        case Left:
            rect.moveTo(m_appAreaWidget->pos());
            break;
        }
    } else {
        switch (m_position) {
        case Top:
        case Bottom :
            rect.moveCenter(this->rect().center());
            if (rect.right() > m_appAreaWidget->geometry().right()) {
                rect.moveRight(m_appAreaWidget->geometry().right());
            }
            if (rect.left() < m_appAreaWidget->geometry().left()) {
                rect.moveLeft(m_appAreaWidget->geometry().left());
            }
            break;
        case Right:
        case Left:
            rect.moveCenter(this->rect().center());
            if (rect.bottom() > m_appAreaWidget->geometry().bottom()) {
                rect.moveBottom(m_appAreaWidget->geometry().bottom());
            }
            if (rect.top() < m_appAreaWidget->geometry().top()) {
                rect.moveTop(m_appAreaWidget->geometry().top());
            }
            break;
        }
    }

    m_appAreaSonWidget->move(rect.x(), rect.y());
}

/**通知布局器，控件已发生变化，需要重新设置几何位置
 * @brief MainPanelControl::updatePluginsLayout
 */
void MainPanelControl::updatePluginsLayout()
{
    for (int i = 0; i < m_pluginLayout->count(); ++i) {
        QLayout *layout = m_pluginLayout->itemAt(i)->layout();
        if (layout) {
            PluginsItem *pItem = static_cast<PluginsItem *>(layout->itemAt(0)->widget());
            if (pItem && pItem->sizeHint().width() != -1) {
                pItem->updateGeometry();
            }
        }
    }

    // 重新设置几何布局后，需要更新布局
    resizeDockIcon();
}

void MainPanelControl::setToggleDesktopInterval(int ms)
{
    if (m_desktopWidget)
        m_desktopWidget->setToggleDesktopInterval(ms);
}

void MainPanelControl::hideOverflowPopupWindow()
{
    m_overflowItem->hidePopUpWindow();
}

void MainPanelControl::itemUpdated(DockItem *item)
{
    item->updateGeometry();
    resizeDockIcon();
}

void MainPanelControl::paintEvent(QPaintEvent *event)
{
    Q_UNUSED(event);

    QPainter painter(this);
    painter.setRenderHints(QPainter::SmoothPixmapTransform | QPainter::Antialiasing);
    QColor color;
    if (DGuiApplicationHelper::instance()->themeType() == DGuiApplicationHelper::LightType) {
        color = Qt::black;
        painter.setOpacity(0.5);
    } else {
        color = Qt::white;
        painter.setOpacity(0.1);
    }

    painter.fillRect(m_fixedSpliter->geometry(), color);
    painter.fillRect(m_appSpliter->geometry(), color);
    if (m_overflowItem->isVisible())
        painter.fillRect(m_overflowSpliter->geometry(), m_trayAreaLayout->count() > 0 ? color : Qt::transparent);
    else
        painter.fillRect(m_appOverflowWidget->geometry(), m_trayAreaLayout->count() > 0 ? color : Qt::transparent);
    painter.fillRect(m_traySpliter->geometry(), m_pluginLayout->count() > 0 ? color : Qt::transparent);
}

/**重新计算任务栏上应用图标、插件图标的大小，并设置
 * @brief MainPanelControl::resizeDockIcon
 */
void MainPanelControl::resizeDockIcon()
{
    // 总宽度
    int totalLength = ((m_position == Position::Top) || (m_position == Position::Bottom)) ? width() : height();

    // 减去托盘间隔区域
    if (m_tray) {
        totalLength -= (m_tray->trayVisibleItemCount() + 1) * 10;
    }

    // 减去4个分割线的宽度
    totalLength -= 4 * SPLITER_SIZE;

    // 减去显示桌面图标宽度
    totalLength -= ((m_position == Position::Top) || (m_position == Position::Bottom)) ? m_desktopWidget->width() :
                                                                                         m_desktopWidget->height();

    int pluginItemCount = 0;
    int calcPluginItemCount = 0;

    // 因为日期时间大小和其他插件大小有异，为了设置边距，在各插件中增加了一层布局
    // 因此需要通过多一层布局来获取各插件
    for (int i = 0; i < m_pluginLayout->count(); ++ i) {
        QLayout *layout = m_pluginLayout->itemAt(i)->layout();
        if (layout) {
            PluginsItem *w = static_cast<PluginsItem *>(layout->itemAt(0)->widget());
            if (w) {
                // 如果插件大小由自己决定,则不参与计算需要减去其宽度,其他插件则需要参与计算并计数
                if ((m_position == Position::Top || m_position == Position::Bottom) && (w->sizeHint().height() != -1)) {
                    totalLength -= w->width();
                } else if ((m_position == Position::Top || m_position == Position::Bottom) && (w->sizeHint().width() != -1)) {
                    totalLength -= w->height();
                } else {
                    calcPluginItemCount++;
                }

                // 所有插件个数,用于计算插件之间的间隔之和
                pluginItemCount++;
            }
        }
    }

    // 减去插件间隔大小, 只有一个插件或没有插件都是间隔20,2个或以上每多一个插件多间隔10
    if (pluginItemCount > 1)
        totalLength -= (pluginItemCount + 1) * 10;
    else
        totalLength -= 20;

    if (totalLength < 0)
        return;

    // 参与计算的插件的个数包含托盘和插件
    int pluginCount = m_tray ? m_tray->trayVisibleItemCount() + calcPluginItemCount : calcPluginItemCount;

    int iconCount = m_fixedAreaLayout->count() + m_appAreaSonLayout->count() + pluginCount;
    if (iconCount <= 0)
        return;

    // Icon is width or height
    int iconSize = ((m_position == Position::Top) || (m_position == Position::Bottom)) ? height() : width();

    // 计算插件图标的最大或最小值
    int trayItemSize = qBound(20, iconSize, 40);
    if ((m_position == Position::Top) || (m_position == Position::Bottom)) {
        trayItemSize = qMin(trayItemSize,height());
        trayItemSize = std::min(trayItemSize, height() - 20);
    } else {
        trayItemSize = qMin(trayItemSize, width());
        trayItemSize = std::min(trayItemSize, width() - 20);
    }

    if (trayItemSize < 20)
        trayItemSize = 20;

    // 减去插件图标的大小后重新计算固定图标和应用图标的平均大小
    totalLength -= trayItemSize * pluginCount;

    // caculate twice is to get real count of overflow apps
    // and get real count apps on
    auto getResizeMessage = [totalLength, this] (int itemSize, int maxCount) -> ResizeMessage {
        bool showOverflow = false;
        int index = 0;
        int overflowAppCount = 0;
        for (auto child : DockItemManager::instance()->itemList()) {
            if (!child.isNull() && child->itemType() == DockItem::ItemType::App) {
                if (index < maxCount) {
                    index += 1;
                } else {
                    overflowAppCount += 1;
                }
            }
        }

        // INFO: if all count <= 3, or overflowAppCount < 3 just resize it
        // 计算溢出区域时候还要额外计算一个驻留的多于的icon, 那么小于3，应该是
        // overflowAppCount + 1 < 3 ==> overflowAppCount < 2
        if (index + overflowAppCount <= 3 || overflowAppCount < 2) {
            int count = index + overflowAppCount + m_fixedAreaLayout->count();
            if (count == 0) {
                return ResizeMessage { 0, 0 , false };
            }
            int realLen = totalLength / count;
            itemSize = qMin(itemSize, realLen);
            return ResizeMessage { itemSize, index + overflowAppCount, false };
        }

        if (overflowAppCount != 0) {
            showOverflow = true;
            if (maxCount >= 1) {
                // INFO: this meens need to put one icon
                maxCount -= 1;
                // 最后重新计算一下区域
                itemSize = qMin(totalLength / (maxCount + 2 + m_fixedAreaLayout->count()), itemSize);
            } else {
                // 如果有多出区域，只能回到等比例缩小方法了
                maxCount = 0;
                itemSize = totalLength / (2 + m_fixedAreaLayout->count());
            }
        }
        return ResizeMessage { itemSize, maxCount, showOverflow };
    };

    if ((m_position == Position::Top) || (m_position == Position::Bottom)) {
        // realheight
        // 任务栏上下间距2px
        int realHeight = qMax(20, height() - 4);
        totalLength -= m_fixedAreaLayout->count() * realHeight;
        int appIconCount = totalLength / realHeight;
        // 总是有余数，将当前应用图标尺寸的3/4作为最低的临界的数值，> 这个数值就额外显示一个，并且使用resize的方法，将所有显示的app挤压显示在dock上
        if (totalLength % realHeight > realHeight * 0.75) {
            appIconCount += 1;
        }
        ResizeMessage realAppIconCount = getResizeMessage(realHeight, appIconCount);
        calcuDockIconSize(
            realAppIconCount.realAppItemSize,
            realAppIconCount.maxCount,
            realAppIconCount.showOverflow,
            trayItemSize
        );
    } else {
        int realWidth = qMax(20, width() - 4);
        totalLength -= m_fixedAreaLayout->count() * realWidth;
        int appIconCount = totalLength  / realWidth;
        // 同上
        if (totalLength % realWidth > realWidth * 0.75) {
            appIconCount += 1;
        }
        ResizeMessage realAppIconCount = getResizeMessage(realWidth, appIconCount);
        calcuDockIconSize(
            realAppIconCount.realAppItemSize,
            realAppIconCount.maxCount,
            realAppIconCount.showOverflow,
            trayItemSize
        );
    }
}

void MainPanelControl::calcuDockIconSize(int appItemSize, int maxCount, bool showOverflow, int traySize)
{
    bool showOverflowChanged = (showOverflow != m_showOverflow);
    // 在显示溢出区域情况下，m_appOverflowLayout 必须要有2个item,如果只有一个，说明
    // active的那个被移除了，判断为需要刷新
    bool overflowStayAppRemoved = showOverflow && m_appOverflowLayout->count() != 2;
    bool needUpdated = (m_appAreaSonLayout->count() != maxCount);
    // update infomation
    m_maxAppStayCount = maxCount;
    m_showOverflow = showOverflow;
    m_appItemSize = appItemSize;
    // set origin one size
    for (int i = 0; i < m_fixedAreaLayout->count(); ++i) {
        m_fixedAreaLayout->itemAt(i)->widget()->setFixedSize(appItemSize, appItemSize);
    }
    m_overflowItem->setFixedSize(appItemSize, appItemSize);
    if (m_overflowRecentlyUsedApp) {
        m_overflowRecentlyUsedApp->setFixedSize(appItemSize, appItemSize);
    }

    // 计算溢出区域的弹出窗口大小, 如果有溢出区域
    if (showOverflow) {
        setOverflowPopUpSize(m_overflowItem->itemCountOnZone());
    }

    // resize All AppItem
    auto children = DockItemManager::instance()->itemList();
    for (auto child :children) {
        if (!child.isNull() && child->itemType() == DockItem::ItemType::App) {
            child->setFixedSize(appItemSize, appItemSize);
        }
    }

    m_overflowItem->refreshIcon();
    if (m_position == Dock::Position::Top || m_position == Dock::Position::Bottom) {
        m_fixedSpliter->setFixedSize(SPLITER_SIZE, int(appItemSize * 0.6));
        m_appSpliter->setFixedSize(SPLITER_SIZE, int(appItemSize * 0.6));
        m_traySpliter->setFixedSize(SPLITER_SIZE, int(appItemSize * 0.5));
        m_overflowSpliter->setFixedSize(SPLITER_SIZE, int(appItemSize * 0.6));
    } else {
        m_fixedSpliter->setFixedSize(int(appItemSize * 0.6), SPLITER_SIZE);
        m_appSpliter->setFixedSize(int(appItemSize * 0.6), SPLITER_SIZE);
        m_traySpliter->setFixedSize(int(appItemSize * 0.5), SPLITER_SIZE);
        m_overflowSpliter->setFixedSize(int(appItemSize * 0.6), SPLITER_SIZE);
    }

    // 如果一个也不能摆放，设置隐藏
    if (maxCount == 0) {
        m_appSpliter->setFixedSize(0, 0);
    }

    // 不显示时候隐藏
    if (!showOverflow) {
        m_overflowSpliter->setFixedSize(0, 0);
    }

    if (m_tray) {
        m_tray->centralWidget()->setProperty("iconSize", traySize);
    }

    //因为日期时间大小和其他插件大小有异，为了设置边距，在各插件中增加了一层布局
    //因此需要通过多一层布局来获取各插件
    if ((m_position == Position::Top) || (m_position == Position::Bottom)) {
        // 三方插件
        for (int i = 0; i < m_pluginLayout->count(); ++ i) {
            QLayout *layout = m_pluginLayout->itemAt(i)->layout();
            if (!layout || !layout->itemAt(0))
                continue;

            PluginsItem *pItem = static_cast<PluginsItem *>(layout->itemAt(0)->widget());
            if (!pItem)
                continue;

            if (pItem->sizeHint().height() == -1) {
                pItem->setFixedSize(traySize, traySize);
            } else if (pItem->sizeHint().height() > height()) {
                pItem->resize(pItem->width(), height());
            }
        }
    } else {
        // 三方插件
        for (int i = 0; i < m_pluginLayout->count(); ++ i) {
            QLayout *layout =  m_pluginLayout->itemAt(i)->layout();
            if (!layout || !(qobject_cast<PluginsItem *>(layout->itemAt(0)->widget())))
                continue;

            PluginsItem *pItem = static_cast<PluginsItem *>(layout->itemAt(0)->widget());
            if (pItem->sizeHint().width() == -1) {
                pItem->setFixedSize(traySize, traySize);
            } else if (pItem->sizeHint().width() > width()) {
                pItem->resize(width(), pItem->height());
            }
        }
    }

    int appTopAndBottomMargin = 0;
    int appLeftAndRightMargin = 0;

    int trayTopAndBottomMargin = 0;
    int trayLeftAndRightMargin = 0;

    if ((m_position == Position::Top) || (m_position == Position::Bottom)) {
        appTopAndBottomMargin = (m_fixedAreaWidget->height() - appItemSize) / 2;
        trayTopAndBottomMargin = (m_trayAreaWidget->height() - traySize) / 2;
    } else {
        appLeftAndRightMargin = (m_fixedAreaWidget->width() - appItemSize) / 2;
        trayLeftAndRightMargin = (m_trayAreaWidget->width() - traySize) / 2;
    }

    m_fixedAreaLayout->setContentsMargins(appLeftAndRightMargin, appTopAndBottomMargin, appLeftAndRightMargin, appTopAndBottomMargin);
    m_appOverflowLayout->setContentsMargins(appLeftAndRightMargin, appTopAndBottomMargin, appLeftAndRightMargin, appTopAndBottomMargin);
    m_appAreaSonLayout->setContentsMargins(appLeftAndRightMargin, appTopAndBottomMargin, appLeftAndRightMargin, appTopAndBottomMargin);
    m_trayAreaLayout->setContentsMargins(trayLeftAndRightMargin, trayTopAndBottomMargin, trayLeftAndRightMargin, trayTopAndBottomMargin);
    // 因为日期时间插件或第三方插件声明自定义大小
    // 而不对自定义大小插件设置边距
    for (int i = 0; i < m_pluginLayout->count(); ++i) {
        QLayout *layout = m_pluginLayout->itemAt(i)->layout();
        if (!layout || !layout->itemAt(0))
            continue;

        PluginsItem *pItem = static_cast<PluginsItem *>(layout->itemAt(0)->widget());
        if (pItem && pItem->pluginSizePolicy() != PluginsItemInterface::Custom) {
            layout->setContentsMargins(trayLeftAndRightMargin, trayTopAndBottomMargin, trayLeftAndRightMargin, trayTopAndBottomMargin);
        }
    }

    if (showOverflowChanged || needUpdated || overflowStayAppRemoved) {
        emit requestUpdateLayout(true);
    }
}

/**重新排布布局，主要是Item的分类
 * @brief mainpanelcontrol::updateLayout
 */
void MainPanelControl::updateLayout(bool needAdd, bool forceShow)
{
    if (m_isDraging) {
        return;
    }

    auto children = DockItemManager::instance()->itemList();
    int index = 0;
    int overflowCount = 0;
    int itemIndex = 0;
    int stayIndex = 0;
    int overflowShowIndex = -1;
    bool hasAttention = false;
    QList<QString> overflowAppId; // id与index一一对应
    QList<int> overflowAppIndex;
    for (auto child : children) {
        if (!child.isNull() && child->itemType() == DockItem::ItemType::App) {
            if (index < m_maxAppStayCount) {
                child->setParent(m_appAreaSonWidget);
                if (forceShow)
                    child->setVisible(true);
                if (needAdd) {
                    addAppAreaItem(-1, child);
                }
                qobject_cast<AppItem *>(child.data())->setTipSwitch(true);
                qobject_cast<AppItem *>(child.data())->setOverflow(false);
                stayIndex++;
            } else {
                if (needAdd) {
                    m_overflowItem->addItem(child);
                }
                overflowCount++;
                qobject_cast<AppItem *>(child.data())->setOverflow(true);
                // check if there is attention, for example ,message
                if (qobject_cast<AppItem *>(child.data())->hasAttention()) {
                    hasAttention = true;
                    overflowShowIndex = itemIndex;
                }
                if (!hasAttention && qobject_cast<AppItem *>(child.data())->isActive()) {
                    overflowShowIndex = itemIndex;
                }
                overflowAppId.append(qobject_cast<AppItem *>(child.data())->appId());
                overflowAppIndex.append(index+1);
            }
            index++;
        }
        itemIndex++;
    }

    // 不需要溢出区
    if (!m_showOverflow) {
        m_overflowAppOpenRecord.clear();
        m_overflowItem->setVisible(false);
        if (m_overflowRecentlyUsedApp) {
            m_overflowRecentlyUsedApp->setVisible(false);
            delete  m_overflowRecentlyUsedApp;
            m_overflowRecentlyUsedApp = nullptr;
        }
        return;
    }

    m_overflowItem->setVisible(true);
    m_overflowItem->setFixedSize(m_appItemSize, m_appItemSize);

    // 处理当没有打开记录时的最近使用区域显示逻辑
    // 优先显示靠右的最小化应用，其次显示最右边的应用
    auto handleEmptyRecord = [this, children, index, &overflowShowIndex, overflowAppIndex] {
        for (int i = overflowAppIndex.size()-1; i >= 0; --i) {
            if (qobject_cast<AppItem *>(children[overflowAppIndex.at(i)].data())->getAppItemWindowId() != 0) {
                overflowShowIndex = overflowAppIndex.at(i);
                break;
            }
        }
        overflowShowIndex == -1 ? overflowShowIndex = index : overflowShowIndex;
        updateOverflowRecentlyUsedApp(qobject_cast<AppItem *>(children[overflowShowIndex].data()));
    };

    // 没有活跃app记录
    if (overflowShowIndex == -1) {
        if (m_overflowAppOpenRecord.isEmpty()) {
            handleEmptyRecord();
        } else {
            // 检查并更新打开记录
            // 移除驻留等操作会导致原记录索引变动
            // 拖拽布局导致原记录应用被移出溢出区
            const auto openRecordList = m_overflowAppOpenRecord;
            m_overflowAppOpenRecord.clear();
            int appIdIndex = -1;
            for (int i = 0; i < openRecordList.size(); ++i) {
                appIdIndex = overflowAppId.indexOf(openRecordList.at(i).first);
                if (appIdIndex == -1) {
                    continue;
                }
                m_overflowAppOpenRecord.append(QPair<QString, int>(overflowAppId.at(appIdIndex), overflowAppIndex.at(appIdIndex)));
            }

            if (m_overflowAppOpenRecord.isEmpty()) {
                handleEmptyRecord();
                return;
            }

            // 如果打开记录中，存在最小化的应用，则优先显示最近的最小化应用，没有则使用打开记录的最后一个
            int miniIndex = -1;
            for (int i = m_overflowAppOpenRecord.size()-1; i >= 0; --i) {
                if (qobject_cast<AppItem *>(children[m_overflowAppOpenRecord.at(i).second].data())->getAppItemWindowId() != 0) {
                    miniIndex = m_overflowAppOpenRecord.at(i).second;
                    break;
                }
            }
            miniIndex == -1 ? overflowShowIndex = m_overflowAppOpenRecord.last().second : overflowShowIndex = miniIndex;
            updateOverflowRecentlyUsedApp(qobject_cast<AppItem *>(children[overflowShowIndex].data()));
        }
    } else {
        updateOverflowRecentlyUsedApp(qobject_cast<AppItem *>(children[overflowShowIndex].data()));
    }

    setOverflowPopUpSize(overflowCount);
}

void MainPanelControl::updateOverflowRecentlyUsedApp(const AppItem *app)
{
    if (!app) {
        qWarning() << "The recently used application is nullptr !!! ";
        return;
    }

    if (!m_showOverflow) {
        return;
    }

    auto newRecentlyUsedApp = [app, this] {
        if (m_overflowRecentlyUsedApp) {
            delete  m_overflowRecentlyUsedApp;
            m_overflowRecentlyUsedApp = nullptr;
        }
        m_overflowRecentlyUsedApp = new AppItem(app->getItemAppSettings(), app->getItemActiveAppSettings(), app->getItemDockedAppSettings(), app->getItemEntry());
        connect(m_overflowRecentlyUsedApp, &AppItem::requestActivateWindow, m_overflowAppInter, &DBusDock::ActivateWindow, Qt::QueuedConnection);
        connect(m_overflowRecentlyUsedApp, &AppItem::requestPreviewWindow, m_overflowAppInter, &DBusDock::PreviewWindow);
        connect(m_overflowRecentlyUsedApp, &AppItem::requestCancelPreview, m_overflowAppInter, &DBusDock::CancelPreviewWindow);
        connect(m_overflowRecentlyUsedApp, &AppItem::requestWindowAutoHide, this, &MainPanelControl::requestOverflowRecentlyUsedAppAutoHide);

        m_overflowRecentlyUsedApp->setVisible(true);
        m_overflowRecentlyUsedApp->setFixedSize(m_appItemSize, m_appItemSize);
        m_overflowRecentlyUsedApp->refreshIcon();
        m_appOverflowLayout->insertWidget(0, m_overflowRecentlyUsedApp);
    };

    if (!m_overflowRecentlyUsedApp) {
        newRecentlyUsedApp();
        return;
    }

    // 有变动才更新
    if (m_overflowRecentlyUsedApp->appId() != app->appId()) {
        newRecentlyUsedApp();
    } else {
        m_overflowRecentlyUsedApp->setVisible(true);
        m_overflowRecentlyUsedApp->setFixedSize(m_appItemSize, m_appItemSize);
        m_overflowRecentlyUsedApp->refreshIcon();
    }
}

void MainPanelControl::setOverflowPopUpSize(int itemCount)
{
    int maxWidth = width() - m_overflowPopMargin * 2;
    int maxHeight = height() - m_overflowPopMargin * 2;
    int areaMaxLen = itemCount * m_appItemSize;
    int realwidth = qMin(maxWidth, areaMaxLen);
    int realheight = qMin(maxHeight, areaMaxLen);

    QPoint panelPos = this->parent() ? qobject_cast<QWidget *>(this->parent())->pos() : pos();
    switch (m_position) {
    case Dock::Left:
    case Dock::Right:
        m_overflowItem->setPopUpSize(width(), realheight, size(), panelPos, m_appItemSize);
        break;
    case Dock::Top:
    case Dock::Bottom:
        m_overflowItem->setPopUpSize(realwidth, height(), size(), panelPos, m_appItemSize);
        break;
    }
}

/**时尚模式没有‘显示桌面’区域
 * @brief MainPanelControl::resizeDesktopWidget
 */
void MainPanelControl::resizeDesktopWidget()
{
    if (m_position == Position::Right || m_position == Position::Left)
        m_desktopWidget->setFixedSize(width(), DESKTOP_SIZE);
    else
        m_desktopWidget->setFixedSize(DESKTOP_SIZE, height());

    if (DisplayMode::Fashion == m_dislayMode) {
        m_desktopWidget->setFixedSize(0, 0);
        m_overflowPopMargin = 0;
    } else {
        m_overflowPopMargin = 10;
    }
    m_overflowItem->setLayoutMargin(m_overflowPopMargin);
}

/**
 * @brief MainPanelControl::checkNeedShowDesktop 根据窗管提供接口（当前是否显示的桌面），提供鼠标
 * 移入 显示桌面窗口 区域时，是否需要显示桌面判断依据
 * @return 窗管返回 当前是桌面 或 窗管接口查询失败 返回false，否则true
 */
bool MainPanelControl::checkNeedShowDesktop()
{
    QDBusInterface wmInter("com.deepin.wm", "/com/deepin/wm", "com.deepin.wm");
    QList<QVariant> argumentList;
    QDBusMessage reply = wmInter.callWithArgumentList(QDBus::Block, QStringLiteral("GetIsShowDesktop"), argumentList);
    if (reply.type() == QDBusMessage::ReplyMessage && reply.arguments().count() == 1) {
        return !reply.arguments().at(0).toBool();
    }

    qDebug() << "wm call GetIsShowDesktop fail, res:" << reply.type();
    return false;
}

/**
 * @brief MainWindow::appIsOnDock 判断指定的应用（驻留和运行显示在任务栏的所有应用）是否在任务栏上
 * @param appDesktop 应用的desktop文件的完整路径
 * @return true: 在任务栏；false: 不在任务栏
 */
bool MainPanelControl::appIsOnDock(const QString &appDesktop)
{
    return DockItemManager::instance()->appIsOnDock(appDesktop);
}

void MainPanelControl::setKwinAppItemMinimizedGeometry(DockItem *item, const QRect r)
{
    // 只处理wayland窗口
    if (!Utils::IS_WAYLAND_DISPLAY) {
        return;
    }

    // 对于wayland应用窗口，需要将窗口的位置传递给窗管，用于设置窗口最小化的位置
    auto w = static_cast<QWidget *>(parent());
    const auto ratio = devicePixelRatioF();
    if (item->itemType() == DockItem::App && w->windowHandle() && w->windowHandle()->handle()) {
        AppItem * appItem = dynamic_cast<AppItem *>(item);
        if (!appItem) {
            qWarning() << "invalid item";
            return;
        }

        QList<QVariant> varList = {0, 0, 0, 0, 0};
        varList[0] = appItem->getActiveWindowId();
        // 此处传给窗管的坐标需要去掉缩放因子，防止最小化窗口位置错误
        varList[1] = (r.x() - w->geometry().x()) * ratio;
        varList[2] = (r.y() - w->geometry().y()) * ratio;
        varList[3] = r.width();
        varList[4] = r.height();
        QGuiApplication::platformNativeInterface()->setWindowProperty(w->windowHandle()->handle(), "_d_dwayland_dock-appitem-geometry", varList);
    }
}
